from ursina import *
from pprint import pprint


# override the built in round to easily round Vec3 and tuples/lists
builtin_round = round
def round(value, decimals=0):
    if isinstance(value, Vec3):
        return Vec3(builtin_round(value[0],decimals), builtin_round(value[1],decimals), builtin_round(value[2],decimals))

    if isinstance(value, list):
        return [builtin_round(e,decimals) for e in value]

    if isinstance(value, tuple):
        return tuple(builtin_round(e,decimals) for e in value)

    return builtin_round(value, decimals)


app = Ursina()

# make hexagon
w = 3.46
verts = list()
points = list()

# bottom part of the hexagon
for y in range(1, 6):
    for x in range(y):
        points.append(Vec3((x+.5-(y/2))*w, y-1, 0))
        verts.extend((
            Vec3((x+.5-(y/2))*w, y-1, 0), Vec3((x+1-(y/2))*w, y, 0), Vec3((x+.5-(y/2))*w, y+1, 0),
            Vec3((x+.5-(y/2))*w, y-1, 0), Vec3((x+.5-(y/2))*w, y+1, 0), Vec3((x-(y/2))*w, y, 0),
        ))

# middle part of the hexagon
for y in range(1, 10):
    length = 6
    offset = .5
    if y%2 == 0:
        length -= 1
        offset = 1

    for x in range(length):
        points.append(Vec3(((x+offset-(6/2))*w), y+4, 0))
        if x == 0 and y < 10 and y%2 != 0:
            verts.extend((
                Vec3(((x+offset-(6/2))*w), y+4, 0), Vec3(((x+offset+.5-(6/2))*w), y+5, 0), Vec3(((x+offset-(6/2))*w), y+6, 0),
            ))
        elif x == length-1 and y < 10 and y%2 != 0:
            verts.extend((
            Vec3(((x+offset-(6/2))*w), y+4, 0), Vec3(((x+offset-(6/2))*w), y+6, 0), Vec3(((x+offset-.5-(6/2))*w), y+5, 0),
            ))

        else:
            verts.extend((
                Vec3(((x+offset-(6/2))*w), y+4, 0), Vec3(((x+offset+.5-(6/2))*w), y+5, 0), Vec3(((x+offset-(6/2))*w), y+6, 0),
                Vec3(((x+offset-(6/2))*w), y+4, 0), Vec3(((x+offset-(6/2))*w), y+6, 0), Vec3(((x+offset-.5-(6/2))*w), y+5, 0),
            ))

# top part of the hexagon
for y in range(5, 0, -1):
    for x in range(y):
        points.append(Vec3((x+.5-(y/2))*w, 19-y, 0))
        verts.extend((
            Vec3((x+.5-(y/2))*w, 19-y, 0), Vec3((x+1-(y/2))*w, 20-y, 0), Vec3((x+.5-(y/2))*w, 21-y, 0),
            Vec3((x+.5-(y/2))*w, 19-y, 0), Vec3((x+.5-(y/2))*w, 21-y, 0), Vec3((x-(y/2))*w, 20-y, 0),
        ))

outline_lines = (
    (Vec3(0*w, 0, 0)),
    Vec3(2.5*w, 5, 0),
    Vec3(2.5*w, 15, 0),
    Vec3(0*w, 20, 0),
    Vec3(-2.5*w, 15, 0),
    Vec3(-2.5*w, 5, 0),
    )
outline_lines = (
    (outline_lines[0], outline_lines[1]),
    (outline_lines[1], outline_lines[2]),
    (outline_lines[2], outline_lines[3]),
    (outline_lines[3], outline_lines[4]),
    (outline_lines[4], outline_lines[5]),
    (outline_lines[5], outline_lines[0]),
)
outline_points = list()
for line in outline_lines:
    segments = 10
    for j in range(segments):
        p = lerp(line[0], line[1], j/segments)
        p = tuple(round(p, 3))
        outline_points.append(p)


outline_mesh = Entity(model=Mesh(vertices=outline_points, mode='point', thickness=3), color=color.cyan, z=-2)
model = Mesh(verts, mode='triangle')
hexagon_points = Entity(model=Mesh(vertices=points, mode='point', thickness=5))
hexagon = Entity(model=model, color=color.white33)


new_verts = list()
subdivided_quads = list()
new_mesh = Entity(model=Mesh(vertices=new_verts, thickness=10, mode='triangle'), color=color.orange, z=-.1)
new_mesh_2 = Entity(model=Mesh(vertices=subdivided_quads, mode='triangle'), color=color.dark_gray.tint(-.025), z=-.2)
# new_mesh_3 = Entity(model=Mesh(vertices=subdivided_quads, mode='line'), color=color.lime, z=-.3)

tris = list()
for i in range(0, len(verts), 3):
    tris.append((verts[i+0], verts[i+1], verts[i+2]))

# pprint(tris)
points = dict()
random.seed(17)
grid_quads = list()
# random.seed(26)

def connect():
    # print(len(tris))
    current_tri = random.choice(tris)

    connections = list()
    for t in tris:
        if len(set(tuple(current_tri) + tuple(t))) == 4: # get pairs with only 4 unique points, they are valid neigbors
            connections.append(t)

    if not connections:
        print('no valid connections found')
        return False

    random_neighbor = random.choice(connections)
    current_center = (current_tri[0] + current_tri[1] + current_tri[2]) / 3
    neighbor_center = (random_neighbor[0] + random_neighbor[1] + random_neighbor[2]) / 3
    center = (current_center + neighbor_center) / 2

    quad = set(current_tri + random_neighbor)

    if current_center[1] == neighbor_center[1]:  # flat ones <>
        s, n = [e for e in quad if not e[1] == center[1]]
        if s[1] > n[1]:
            s, n = n, s
        w, e = [e for e in quad if not e in (s,n)]
        if w[0] > e[0]:
            w, e = e, w

    elif current_center[0] < neighbor_center[0]:
        if current_center[1] < neighbor_center[1]:  # right up
            s,e,n,w = current_tri[0], current_tri[1], random_neighbor[1], current_tri[2]
        else:   # right down
            s,e,n,w = current_tri[0], random_neighbor[0], current_tri[1], current_tri[2]
    else:   # left up
        if current_center[1] < neighbor_center[1]:
            s,e,n,w = current_tri[0], current_tri[1], random_neighbor[2], current_tri[2]
        else:   # left down
            s,e,n,w = current_tri[0], current_tri[1], current_tri[2], random_neighbor[0]

    se,ne,nw,sw = (s+e)/2, (n+e)/2, (n+w)/2, (s+w)/2

    cc_shape = (s,se,e,ne,n,nw,w,sw) # points in counter clockwise order
    cc_shape = [round(e, 3) for e in cc_shape]
    betweens = (cc_shape[1], cc_shape[3], cc_shape[5], cc_shape[7])

    rounded_center = round(center, 3)
    if not str(rounded_center) in points:
        points[str(rounded_center)] = betweens

    for i, p in enumerate(cc_shape):
        name = str(p)
        if not name in points:
            points[name] = list()

        prev = cc_shape[i-1]
        if i == len(cc_shape)-1:
            next = cc_shape[0]
        else:
            next = cc_shape[i+1]

        points[name].append(Vec3(prev))
        points[name].append(Vec3(next))

        if p in betweens:
            points[name].append(Vec3(*rounded_center))


    # visuals
    quads = ((center,se,e,ne), (center,ne,n,nw), (center,nw,w,sw), (center,sw,s,se))
    grid_quads.extend(quads)
    for q in quads:
        a,b,c,d = q
        # c = (a+b+c+d) / 4
        # a,b,c,d = [lerp(c, e, .95) for e in q]
        new_mesh_2.model.vertices.extend((a,b,c, a,c,d))

    new_mesh_2.model.generate()
    # new_verts.extend(current_tri)
    # new_verts.extend(random_neighbor)
    new_mesh.model.vertices.extend([lerp(center, v, .9) for v in current_tri])
    new_mesh.model.vertices.extend([lerp(center, v, .9) for v in random_neighbor])
    new_mesh.model.generate()

    tris.remove(current_tri)
    tris.remove(random_neighbor)
    return True


def subdivide_triangles():
    for current_tri in tris:
        center = (current_tri[0] + current_tri[1] + current_tri[2]) / 3

        a, b, c = current_tri
        ab, bc, ca = (a+b)/2, (b+c)/2, (c+a)/2

        cc_shape = (a,ab,b,bc,c,ca) # points in counter clockwise order
        cc_shape = [round(e,3) for e in cc_shape]
        betweens = (cc_shape[1], cc_shape[3], cc_shape[5])
        rounded_center = round(center, 3)
        if not str(rounded_center) in points:
            points[str(rounded_center)] = betweens

        for i, p in enumerate(cc_shape):
            name = str(p)
            if not name in points:
                points[name] = list()

            prev = cc_shape[i-1]
            if i == len(cc_shape)-1:
                next = cc_shape[0]
            else:
                next = cc_shape[i+1]

            points[name].append(Vec3(prev))
            points[name].append(Vec3(next))

            if p in betweens:
                points[name].append(Vec3(*rounded_center))

        # visuals
        quads = ((center,ab,b,bc), (center,bc,c,ca), (center,ca,a,ab))
        for q in quads:
            a,b,c,d = q
            c = (a+b+c+d) / 4
            a,b,c,d = [lerp(c, e, .95) for e in q]
            new_mesh_2.model.vertices.extend((a,b,c, a,c,d))

        new_mesh.model.vertices.extend([lerp(center, v, .9) for v in current_tri])

    new_mesh.model.generate()
    new_mesh_2.model.generate()

    tris.remove(current_tri)


def smooth():
    print('smooth')
    quad_ids = list()
    for q in grid_quads:
        quad_ids.append([new_mesh_2.model.vertices.index(pos) for pos in q])
        # for pos in q:
            # print(new_mesh_2.model.vertices.index(pos))
        # print(q)
    # quad_ids = list()
    # for i in range(0, len(new_mesh.model.vertices)-4, 4):
    #     # print('---', [round(new_mesh.model.vertices[i+j], 3) for j in range(4)])
    #     quad = [new_mesh.model.vertices[i+j] for j in range(4)]
    #     quad = [new_mesh.model.vertices.index(pos) for pos in quad]
    #     # for pos in q:
    #     #     print(new_mesh.model.vertices.index(pos))
    #         # print(new_mesh.model.vertices.index[pos])
    #     print('...', quad, type(quad))
    #     quad_ids.append(quad)

    print(max(quad_ids))
    Entity(model=Mesh(new_mesh_2.model.vertices, triangles=quad_ids), color=color.green, z=-5)

    vertices = [eval(e.replace('LVector3f', '')) for e in points.keys()]
    print('len vertices:', len(vertices))
    lines = list()


    for key, value in points.items():
        a = vertices.index(eval(key.replace('LVector3f', '')))
        for pos in value:
            pos = round(tuple(pos),3)
            # print(pos)
            b = vertices.index(pos)
            if not (b,a) in lines:
                lines.append((a,b))

    connections = list()
    for i, v in enumerate(vertices):
        _lines = [l for l in lines if i in l]
        neighbors = list()

        for l in _lines:
            neighbors.extend([e for e in l if not e == i])

        connections.append(tuple(set(neighbors)))

    relaxed_verts = relax(vertices, connections)
    # relaxed_verts = relax(relaxed_verts, connections)
    # render
    new_mesh_2.enabled = False
    new_mesh.enabled = False

    point_mesh = Entity(model=Mesh(vertices=list(), mode='point', thickness=5), z=-1, color=color.orange)

    for i, neighbors in enumerate(connections):
        neighbor_positions = [relaxed_verts[e] for e in neighbors]
        neighbor_positions = [Vec3(*e) for e in neighbor_positions]

        point_mesh.model.vertices.append(relaxed_verts[i])

        lines_mesh = Entity(model=Mesh(vertices=list(), mode='line', thickness=1), color=color.black, z=-1)
        for pos in neighbor_positions:
            lines_mesh.model.vertices.extend((relaxed_verts[i], pos))

        lines_mesh.model.generate()

    point_mesh.model.generate()
    print('finished')


def relax(points, connections):
    relaxed_verts = list()
    for i, neighbors in enumerate(connections):
        neighbor_positions = [points[e] for e in neighbors]
        neighbor_positions = [Vec3(*e) for e in neighbor_positions]

        mean = neighbor_positions[0]
        for np in neighbor_positions[1:]:
            mean += np

        mean /= len(neighbor_positions)


        if points[i] in outline_points:
            # Entity(model='quad', scale=.4, color=color.pink, position=points[i], z=-3)
            relaxed_verts.append(points[i])
            continue

        relaxed_verts.append(mean)

    return relaxed_verts


def connect_randomly():
    misses = 0
    while misses < 16:
        if connect() == False:
            misses += 1

    print('finished')

# # animate
# s = Sequence(Wait(2))
# for i in range(0, len(verts)+3, 3):
#     s.append(Func(setattr, model, 'vertices', verts[:i]))
#     s.append(Func(model.generate))
#     s.append(Wait(.025))
#
# s.start()
# subdivide each quad
misses = 0
def input(key):
    global misses
    if key == 'space' or key == 'space hold':
        connect()


    if key == 'r':
        if misses < 16:
            if connect() == False:
                misses += 1

            connect_randomly()
            connect_randomly()
        # else:
        #     subdivide_triangles()


    if key == 't':
        subdivide_triangles()
        return
        s = Sequence()
        new_mesh_2.model.vertices = ((0,0,0),(0,0,0),(0,0,0))
        new_mesh_2.model.generate()
        for i in range(0, len(subdivided_quads)+6, 6):
            s.append(Func(setattr, new_mesh_2.model, 'vertices', subdivided_quads[:i]))
            s.append(Func(new_mesh_2.model.generate))
            s.append(Wait(.0125))

        s.start()

    if key == 's':
        smooth()


# window.render_mode = 'wireframe'

camera.orthographic = True
camera.fov = 30
camera.y = 10
window.size *= .75
# EditorCamera()

app.run()



from ursina import *


app = Ursina()
from ursina.shaders import normals_shader

desert_scene = load_blender_scene('blender_level_editor_test_scene_2')
# desert_scene.Cube.color=color.red
desert_scene.Cube.shader = normals_shader

EditorCamera()


def update():
    desert_scene.Cube.x += (held_keys['d'] - held_keys['a']) * time.dt * 10
    desert_scene.Cube.rotation_x += 1
    desert_scene.Cube.rotation_y += 1
    desert_scene.Cube.rotation_z += 1

app.run()



import re
import traceback


def camel_to_snake(value):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', value)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def snake_to_camel(value):
    camel = ''
    words = value.split('_')
    for w in words:
        camel += w.title()
    return camel


def multireplace(string, replacements, ignore_case=False):
    """
    Given a string and a dict, replaces occurrences of the dict keys found in the
    string, with their corresponding values. The replacements will occur in "one pass",
    i.e. there should be no clashes.
    :param str string: string to perform replacements on
    :param dict replacements: replacement dictionary {str_to_find: str_to_replace_with}
    :param bool ignore_case: whether to ignore case when looking for matches
    :rtype: str the replaced string
    """
    rep_sorted = sorted(replacements, key=lambda s: len(s[0]), reverse=True)
    rep_escaped = [re.escape(replacement) for replacement in rep_sorted]
    pattern = re.compile("|".join(rep_escaped), re.I if ignore_case else 0)
    return pattern.sub(lambda match: replacements[match.group(0)], string)

def printvar(var):
     print(traceback.extract_stack(limit=2)[0][3][9:][:-1],"=", var)

def print_info(str, *args):
    from ursina import application
    if application.print_info:
        print('info:', str, *args)

def print_warning(str, *args):
    from ursina import application
    if application.print_warnings:
        print('\033[93mwarning:', str, *args, '\033[0m')



if __name__ == '__main__':
    from ursina .ursinastuff import _test
    _test(camel_to_snake('CamelToSnake') == 'camel_to_snake')
    _test(snake_to_camel('snake_to_camel') == 'SnakeToCamel')

    test_var = 1
    printvar(test_var)



from ursina import *



app = Ursina()

head = Entity(model='sphere', scale=1, scale_x=.9)
jaw = Entity(model=Cylinder(12), position=(0,-.5, -.15), scale=(.8,.5,.7))
jaw.model.generate_normals()

nose = Entity(model='cube',)


window.display_mode = 'normals'

EditorCamera()
app.run()



from ursina import *
from cool_scene_transition import CoolSceneTransition



app = Ursina(forced_aspect_ratio=16/9)


def swooshy_text(text, background_color=color.black, scale=1.5):
    pivot = Entity()
    # pivot = Entity(world_parent=camera.ui)
    t = Text(parent=pivot, z=-10, text=text, scale=20, origin=(0,0), background=True)
    t.create_background(padding=(1,.5), radius=.01, color=background_color)

    pivot.rotation_y = -45
    pivot.animate_rotation_y(0, duration=1*scale, curve=curve.in_out_expo)
    pivot.animate_rotation_y(45, duration=1*scale, delay=1.3*scale, curve=curve.in_out_expo)
    t.color = color.clear
    t.animate_color(color.smoke, duration=.5*scale)
    t.animate_color(color.clear, duration=.5*scale, delay=1.3*scale)
    t.background.color = color.clear
    t.background.animate_color(background_color, duration=.5*scale)
    t.background.animate_color(color.clear, duration=.5*scale, delay=1.3*scale)
    t.x = 1
    t.animate_x(-1, duration=2*scale, curve=curve.linear)
    destroy(pivot, delay=2*scale)


# window.fullscreen = True
ursina_green = color.color(122, .57, .69).tint(-.15)
ursina_green = color.azure.tint(0)
import trailer_code_example
window.color = color.black
Text.default_resolution *= 2
Text.default_font = 'VeraMono.ttf'
Texture.default_filtering = 'bilinear'
step = 8.72/8

intro_text = Text("A new <lime>game engine<default> has entered the arena!", origin=(0,0), scale=2, enabled=False)
music = Audio('Reloaded Installer _3.mp3', autoplay=False)

logo = Sprite(name='ursina_splash', parent=camera.ui, texture='ursina_logo', world_z=camera.overlay.z-1, scale=.12, color=color.clear)
bg = Entity(model=Cylinder(16, mode='line', thickness=1), color=color.azure, scale=80, scale_y=16, x=-20, y=-5, enabled=False)
inner_wall = Entity(parent=bg, model=Cylinder(16, mode='line', thickness=1), scale=.25, scale_y=1, color=bg.color)

trailer_code_example.tf.x -= .2

video_parent = Entity(parent=camera.ui, x=-.5*camera.aspect_ratio)
video_parent.world_parent = scene
video_parent.original_position = video_parent.position
video_parent.rotate = True
video_parent.rotation_speed = .01

def reset():
    video_parent.position = video_parent.original_position
    video_parent.rotation_y = 0
video_parent.reset = reset

class Video(Entity):
    def __init__(self, **kwargs):
        super().__init__(
            parent=camera.ui, model='quad', scale_x=camera.aspect_ratio, world_parent=video_parent, enabled=False)

        for key, value in kwargs.items():
            setattr(self, key, value)



one_gameplay =        Video(texture='one_gameplay.mp4')
inventory_gameplay =  Video(texture='inventory_example_gameplay.mp4')
value_of_life =       Video(texture='value_of_life_gameplay.mp4')

voxel_tool =          Video(texture='voxel_tool_gameplay_cut.mp4', position=(.4,-.2))
voxel_tool.scale *= .5
voxel_tool.x = 1.4

hotreloading =        Video(texture='code_hotreloading.mp4')
hotreloading.scale *= 1.05

cool_scene_transition = CoolSceneTransition()
loddefjord = Video(texture='loddefjord.mp4')
# loddefjord.texture.set_play_rate(1.4)
loddefjord.texture.set_play_rate(.85)
loddefjord.texture.stop()

column_graph = Video(texture='column_graph.mp4')
column_graph.texture.set_play_rate(2.9)

mysterious_sphere = Video(texture='mysterious_sphere.mp4')
mysterious_sphere.texture.stop()

autoblock_footage = Video(texture='autoblock_footage.mp4')
autoblock_footage.texture.set_play_rate(3)

# tunnel = Entity(model=Prismatoid(base_shape=Circle(12), path=[Vec3(0,0,i*2) for i in range(8)],mode='line'), scale=10, color=color.azure.tint(-.2), z=-12, enabled=False)
download_text = Text("Get started now!", origin=(0,0), scale=1.5, enabled=False, y=.1)
link =          Text("www.ursinaengine.org", origin=(0,0), scale=2, enabled=False, y=-.0)
music_credit =  Text("Music by LHS", origin=(0,-.5), scale=1, enabled=False, y=-.47)

for movie in (loddefjord, column_graph, mysterious_sphere, autoblock_footage):
    movie.scale *= 1.125
    movie.texture.stop()

for movie in (inventory_gameplay, value_of_life, voxel_tool):
    movie.texture.stop()
    movie.texture.set_loop_count(1)


cumulative_time = 25.29750

s = Sequence(
    Func(music.play),
    Func(setattr, bg, 'enabled', True),
    Func(setattr, intro_text, 'enabled', True), Func(intro_text.appear, speed=.05),
    Wait(step*2),
    Func(intro_text.fade_out),
    Wait(step*.5),

    Func(logo.animate_color, color.white, duration=step),
    Func(logo.animate_scale, logo.scale*1.25, curve=curve.linear, duration=step*8),
    Wait(step*5),
    Func(logo.animate_color, color.clear, curve=curve.linear, duration=step),
    Wait(step*1),

    Func(camera.overlay.blink, color.black, duration=1),
    Wait(.5),
    Func(swooshy_text, "Make Games", background_color=ursina_green),
    Func(setattr, bg, 'enabled', False), Func(video_parent.reset),

    Func(one_gameplay.texture.play), Func(setattr, one_gameplay, 'enabled', True), Wait(step*2.75), Func(setattr, one_gameplay, 'enabled', False), Func(video_parent.reset),
    Func(inventory_gameplay.texture.play), Func(setattr, inventory_gameplay, 'enabled', True), Wait(step*2.25), Func(setattr, inventory_gameplay, 'enabled', False),Func(video_parent.reset),
    Func(setattr, value_of_life, 'enabled', True), Func(value_of_life.texture.play), Wait(step*2.25), Func(setattr, value_of_life, 'enabled', False),

    Func(swooshy_text, "Code in Python", background_color=ursina_green),
    step*2,
    Func(trailer_code_example.tf.text_entity.appear, speed=.0025),
    Func(video_parent.reset),
    Func(setattr, voxel_tool, 'enabled', True), Func(voxel_tool.texture.play),
    step*4,
    Func(camera.overlay.blink, color.black, duration=step),
    step/2,
    Func(setattr, trailer_code_example.tf, 'text', ''),
    Func(trailer_code_example.tf.render),
    Func(setattr, voxel_tool, 'enabled', False),
    Func(setattr, trailer_code_example, 'enabled', False),
    step/2,


    # Func(music.play, cumulative_time),

    Func(setattr, video_parent, 'position', video_parent.original_position), Func(setattr, video_parent, 'rotation_y', 0),

    Func(swooshy_text, "Low Friction Workflow", background_color=ursina_green),
    Func(hotreloading.texture.play), Func(setattr, hotreloading, 'enabled', True),
    Wait(step*5.5),
    Func(cool_scene_transition.play, .9),
    Wait(step*1),
    Func(setattr, hotreloading, 'enabled', False),
    Func(video_parent.reset),
    Func(setattr, video_parent, 'rotate', False),

    Func(setattr, loddefjord, 'enabled', True), Wait(.8),
    Func(swooshy_text, "Free and Open Source", background_color=ursina_green),
    Func(loddefjord.texture.play), Wait(5.5), Func(setattr, loddefjord, 'enabled', False),
    # Func(swooshy_text, "Modern UI", background_color=ursina_green),
    Func(setattr, column_graph, 'enabled', True), Func(column_graph.texture.play), Wait(step*4.7), Func(setattr, column_graph, 'enabled', False),
    Func(setattr, mysterious_sphere, 'enabled', True), Func(mysterious_sphere.texture.play), Wait(step*2.5), Func(setattr, mysterious_sphere, 'enabled', False),
    # Func(swooshy_text, "Procedural Geometry", background_color=ursina_green),
    Func(autoblock_footage.texture.play), Func(setattr, autoblock_footage, 'enabled', True), Wait(7.5),
    Func(camera.overlay.blink, color.black, duration=1), Wait(.5), Func(setattr, autoblock_footage, 'enabled', False),

    Func(setattr, window, 'color', color.color(122, .57, .69).tint(-.3)),
    Func(setattr, download_text, 'enabled', True), Func(download_text.appear, speed=.1),
    2,
    Func(download_text.fade_out, duration=3),
    Func(setattr, link, 'enabled', True), Func(link.appear, speed=.1),
    Func(link.animate_scale, link.scale*1.5, curve=curve.linear, duration=step*8),
    5,
    Func(setattr, music_credit, 'enabled', True),
    Func(setattr, music_credit, 'color', color.clear),
    Func(music_credit.animate_color, color.white, duration=3),
    Func(music.fade_out, duration=8),
)
# window.color = color.lime.tint(-.5)
# window.color = color.color(122, .57, .69).tint(-.3)
window.editor_ui.enabled = False

started = False

def update():

    if started:
        bg.rotation_y += 1
        bg.rotation_x -= .05

        if video_parent.rotate:
            video_parent.rotation_y += video_parent.rotation_speed
            video_parent.x -= .005
        # if tunnel.enabled:
        #     tunnel.z -= .15

def input(key):
    global started

    if not started and key == 'space':
        s.start()
        started = True
        print('start')

camera.fov = 90


app.run()



from ursina import *

app = Ursina()

player = Entity(model='wireframe_cube', color=color.lime)
player.forward_indicator = Entity(parent=player, model=Cone(mode='line'), rotation_x=90, z=.5, color=color.orange)
def input(key):
    if key == 'w':
        player.rotation_y = 0
    if key == 'd':
        player.rotation_y = 90
    if key == 's':
        player.rotation_y = 180
    if key == 'a':
        player.rotation_y = 270


graphics = duplicate(player, parent=scene)
def update():
    graphics.position = player.position
    graphics.rotation_y = lerp_angle(graphics.rotation_y, player.rotation_y, time.dt * 5)

EditorCamera()

app.run()



from ursina import *
from ursina.shaders import basic_lighting_shader
window.vsync = False

app = Ursina()

window.color=color.smoke
board = Entity(model='plane', scale=8, texture='white_cube', texture_scale=(8,8), origin=(-.5,0,-.5))

# make a pawn model by using a procedural Pipe
pawn_model = Pipe(
    base_shape=Circle(16),
    path=[Vec3(0,y,0) for y in [.05, .063, .1, .146, .173, .21, .254, .275, .288, .434, .445, .5, .556, .618, .7, .778, .85, .9, .92]],
    thicknesses=(.632, .58, .64, .64, .56, .4, .384, .350, .25, .230, .4, .45, .2, .35, .43, .434, .38, .26, 0)
    )
pawn_model.generate_normals()


# create pawns
for i in range(8):
    piece = Draggable(
        parent=scene,   # set to scene so it's not on the "ui layer"
        model=copy(pawn_model),   # set the model we create earlier
        color=color.smoke,
        highlight_color=color.lime,
        origin=(-.5,0,-.5),     # offset the model
        plane_direction=(0,1,0),    # make the plane it moves along point upwards
        step=(1,1,1),
        shader=basic_lighting_shader,
        x=i,
        z=1,
        )
    piece.set_shader_input('transform_matrix', piece.getNetTransform().getMat())

    # assign a drop function to the piece which will be called when release the mouse
    def drop():
        piece.position = round(piece.position, 0)
    piece.drop = drop


Cursor()
mouse.visible = False
Sky(texture='sky_sunset')
Entity(model='plane', color=color.light_gray, scale=2000, y=-.01, texture='grass', texture_scale=(100,100))
EditorCamera(rotation_x=30, position=(4,0,4))



app.run()



from ursina import *
from textwrap import dedent
from ursina import DotDict
# WORLD_NORMAL

# void vert():
#     gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
#     texcoord = p3d_MultiTexCoord0;

#     world_normal = normalize(mat3(p3d_ModelMatrix) * p3d_Normal);
#     vertex_world_position = (p3d_ModelMatrix * p3d_Vertex).xyz;
#     vertex_color = p3d_Color;


def sample_texture(texture, uv):
    return sample_bilinear(texture, uv.x*texture.width, uv.y*texture.height)

ursina_to_glsl_map = {
    'sample_bilinear' : 'texture',
    '' : 'sampler2D'
    }

# def vert(vertex_position):
# 'lerp':'mix',
# 'distance':


def fragment(p3d_Texture0:Texture, uv:Vec2, fog_color:Color=color.light_gray, camera_world_position=Vec3.zero, vertex_position=Vec3.zero) -> Color:
    pixel = sample_texture(p3d_Texture0, uv) * Vec4(uv.x, uv.y, uv.x * uv.y, 1.0)
    t = distance(vertex_position, camera_world_position)/400
    result = lerp(pixel, fog_color, t)
    return result


def cpu_shader(original_pixels:Array2D, target_texture:Texture, frag_function, shader_input:dict=None):
    # run shader on cpu

    shader_input = shader_input if shader_input is not None else {}

    for (x,y), pixel in enumerate_2d(original_pixels):
        uv = Vec2(x/original_pixels.width, y/original_pixels.height)
        target_texture.set_pixel(x, y, frag_function(original_pixels, uv, **shader_input))
    target_texture.apply()



def triplanar_mapping_frag(side_texture:Texture, ) -> Color:
    def TriPlanarBlendWeightsConstantOverlap(normal) -> Vec3:
        return normal

    blend:Vec3 = TriPlanarBlendWeightsConstantOverlap(WORLD_NORMAL)

    albedoX:Vec3 = texture(side_texture, vertex_world_position.zy * side_texture_scale).rgb*blend.x
    albedoY:Vec3 = texture(side_texture, vertex_world_position.xz * side_texture_scale).rgb*blend.y
    albedoZ:Vec3 = texture(side_texture, vertex_world_position.xy * side_texture_scale).rgb*blend.z
    a = 1.0
    b:float = 2.0
    b = 2.5

    if WORLD_NORMAL.y > .0:
        albedoY = texture(TEXTURE, vertex_world_position.xz * texture_scale.xy).rgb*blend.y

    triPlanar:Vec3 = (albedoX + albedoY + albedoZ)

    return Vec4(triPlanar.rgb, 1) * VERTEX_COLOR


import ast

def get_variables_types_and_values_in_function(source_code):
    tree = ast.parse(source_code)

    # Find the target function
    target_function = None
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            target_function = node
            # break

            # if not target_function:
            #     raise ValueError(f"Function '{function_name}' not found in the code.")

            # Extract variable names, their associated type hints, and values within the function
            variables_info = {}
            for sub_node in ast.walk(target_function):
                if isinstance(sub_node, ast.Assign):
                    for target in sub_node.targets:
                        if isinstance(target, ast.Name):
                            variable_name = target.id
                            type_hint = None
                            value = None

                            if sub_node.value:
                                value = ast.get_source_segment(source_code, sub_node.value).strip()

                            if isinstance(sub_node.value, ast.AnnAssign):
                                # If there's a type hint in the AnnAssign, use it
                                type_hint = ast.get_source_segment(source_code, sub_node.value.annotation).strip()

                            variables_info[variable_name] = {'type_hint': type_hint, 'value': value, 'function': node.name}
                            print('add:', variable_name, variables_info[variable_name])

            return variables_info



import inspect
import ast
class VarInfo:
    def __init__(self, name='', type_hint=None, default_value=..., comment='', line_number:int=None):
        self.name = name
        self.type_hint = type_hint
        self.default_value = default_value
        self.examples = []
        self.comment = comment
        self.line_number = line_number


class FunctionInfo:
    def __init__(self, name='', type_hint=None, input_args:[VarInfo]=..., comment='', decorators=None, module=None, from_class=None, line_number:int=None):
        self.module = module
        self.from_class = from_class
        self.name = name
        self.type_hint = type_hint
        self.input_args = input_args
        self.comment = comment
        self.decorators = decorators or []
        self.examples = []
        self.line_number = line_number

def extract_function_args(func_node):
    args = func_node.args.args
    defaults = func_node.args.defaults

    num_args = len(args)
    num_defaults = len(defaults)
    default_offset = num_args - num_defaults

    result = []

    for i, arg in enumerate(args):
        var_info = VarInfo()
        var_info.name = arg.arg
        var_info.type_hint = ast.unparse(arg.annotation) if arg.annotation else None

        # Match defaults from the right
        if i >= default_offset:
            default_expr = defaults[i - default_offset]
            var_info.default_value = ast.unparse(default_expr)
        else:
            var_info.default_value = None

        result.append(var_info)

    return result



def frag_to_glsl(func):

    tree = ast.parse(inspect.getsource(func))
    print(ast.dump(tree, indent=4))
    node = tree.body[0]
    func_info = FunctionInfo(
        name = node.name,
        type_hint = ast.unparse(node.returns) if node.returns else Color,
        input_args = extract_function_args(node),
        decorators = [ast.unparse(d) for d in node.decorator_list],
        # comment = inline_comments.get(node.lineno),
        # line_number=node.lineno,
        )
    variables = []

    for action in node.body:
        if isinstance(action, (ast.AnnAssign)):
            var_name = action.target.id
            data_type = action.annotation.id
            print(f'{data_type} {var_name} = {action.value};')
            variables.append(DotDict(name=var_name, type=data_type, value=action.value))

        if isinstance(action, (ast.Assign)):
            var_name = action.targets[0].id
            data_type = '???'
            # datatype = guess_type(action.value)
            variables.append(DotDict(name=var_name, type=data_type, value=action.value))
            print(f'{data_type} {var_name} = {action.value};')
            # print('new var:', 'name:', action.targets[0].id, 'type:', '???', 'value:', action.value)


    result = dedent('''\
        #version 140
        out vec4 result;
        '''
        )

    for var in func_info.input_args:
        result += f'uniform {var.type_hint} {var.name};\n'

    result += f'{func_info.type_hint} {func_info.name}() {{\n'
    result += '}'


    # function_info = inspect.getmembers(func)
    # local_variables = [var for var in function_info if inspect.isframe(var[1])]
    # variable_names = local_variables[0][1].f_locals.keys()
    # print(function_info)
    return result
if __name__ == '__main__':
    glsl_code = frag_to_glsl(fragment)
    print('_________________________')
    print(glsl_code)

    app = Ursina()
    quad = Entity(model='quad', texture='brick', shader_input={'fog_color':color.blue})
    del quad.shader_input['texture_scale']
    del quad.shader_input['texture_offset']
    pixels = quad.texture.pixels

    cpu_shader(pixels, quad.texture, fragment, quad.shader_input)
    def input(key):
        quad.shader_input['camera_world_position'] = camera.world_position
        cpu_shader(pixels, quad.texture, fragment, quad.shader_input)
    EditorCamera()
    app.run()

# from unlit_shader import vert as unlit_shader_vert
# my_shader = wind_shader + unlit_shader + fog_gradient_shader + lit_with_shadows_shader + triplanar_shader
#
# if __name__ == '__main__':
#     import ast



# half rim = 1 - saturate(dot(data.worldViewDir, s.Normal));
# rim = smoothstep(_RimMin, _RimMax, rim);
# #ifdef _USERIMMAP
#     rim *= s.RimIntensityMultiplier;
# #endif
# #ifdef _SUN_COLORED_RIM
#     _RimColor = _LightColor0 * 0.5;
# #endif
# gi.indirect.specular += _RimColor.rgb * rim * _RimColor.a;




from ursina import *

rim_shader = Shader(language=Shader.GLSL,
vertex = '''
#version 430
uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 uv;
in vec3 p3d_Normal;
uniform mat3 p3d_NormalMatrix;
uniform mat4 p3d_ViewMatrix;
uniform mat4 p3d_ViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;
uniform mat4 p3d_ProjectionMatrix;
out vec3 view_normal;

void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
  uv = p3d_MultiTexCoord0;
  view_normal = mat3(p3d_NormalMatrix) * p3d_Normal;
}
''',

fragment='''
#version 430

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 uv;
in vec3 view_normal;
out vec4 color;


void main() {
    vec4 map = texture(p3d_Texture0, uv) * p3d_ColorScale;
    vec3 rgb = 1 - view_normal * 1.;
    color = vec4(map.rgb + rgb.rgb, 1.0);
}

''',
)

if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', shader=rim_shader, rotation_y=45, color=color.azure)

    EditorCamera()
    # def update():
    #     e.rotation_x += 1
    #     e.rotation_y += 1
    #     e.rotation_z += 1

    def input(key):
        if key == 'space':
            if e.shader:
                e.shader = None
            else:
                e.shader = rim_shader

    camera.orthographic = True
    camera.fov = 3
    app.run()



from ursina import *
from ursina.prefabs.primitives import *


# class Arild(Entity):
    # def __init__(self):
    #     super().__init__()

    # def create_gear(self):
class Gear(Draggable):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        print('yo')
        self.parent = scene
        self.model = 'sphere'
        # self.model = Cylinder(12, direction=(0,0,1))

        self.collider = 'sphere'

        self.joint = RedCube(parent=self, y=.5)
        self.joint.scale *= .1
        # self.model=Cylinder(6),

    def input(self, key):
        super().input(key)
        if not self.hovered:
            return

        if key == 'scroll down':
            self.scale *= .9
        if key == 'scroll up':
            self.scale /= .9

    def update(self):
        super().update()
        if self.hovered and held_keys['r']:
            if mouse.left:
                self.rotation_z += 2
            if mouse.right:
                self.rotation_z -= 2

        self.color = color.color(60, self.scale_x/4, 1)


    # Entity(model=Cylinder(6), color=color.color(60,1,1,.3))
    # return Gear()


# def create_rod(self):
class Rod(Entity):
    def __init__(self):
        super().__init__()
        self.parent = scene
        self.model = 'cube'
        self.color = color.red
        self.origin = (0, -.5, 0)
        self.scale = (.2, 1, .2)

    def update(self):
        if hasattr(self, 'position_target'):
            self.world_position = self.position_target.world_position

        if hasattr(self, 'look_at_target'):
            self.look_at(self.look_at_target)

    # return Rod()



if __name__ == '__main__':
    app = Ursina()
    # a = Arild()
    # DebugMenu(target=Arild(), position=(-.7, .3))
    # gear = Gear()
    # rod = Rod()
    # rod.position_target = gear.joint
    # g2 = Gear(scale=(.2,.2,.2), position=(0,1,0))
    # rod.look_at_target = g2
    # look_target = a.create_gear()
    # rod.add_script()
    # cube = Entity(position=(0,1,0), rotation=(0,0,0), scale_y=2, model='cube', origin_y=-.5)

    # slider = Slider(0, 4, default=1, dynamic=True)
    # slider.label.text = 'cube height'
    # def on_value_changed():
    #     cube.scale_y = slider.value
    #     cube.color = color(slider.value*100, 1, 1)
    # slider.on_value_changed = on_value_changed


    line0 = Button(parent=scene, model='cube', scale=(.1,.3,.1), origin=(0,-.5,0), color=color.red)
    line0.end = RedCube(parent=line0, y=1, world_scale=(.1,.1,.1))
    line0.end.scale *= 1.1

    line2 = Entity(position=(1.2,-.5), model='cube', scale=(.1, .1, 1), origin=(0,0,-.5), color=color.orange)
    line2.paused = False
    line2.end = Entity(parent=line2, z=1)

    line1 = Entity(model='cube', scale=(.1,.1,1), origin=(0,0,-.5), color=color.green)
    line1.paused = False
    line1.end = Entity(parent=line1, z=1)

    intersecting_point = Entity(model='sphere', scale=(.1,)*3, color=color.cyan)
    def intersect_update():
        line1.world_position = line0.end.world_position
        line2.look_at(line1.world_position)
        line1.look_at(line2.end)

        for i in range(360):
            dist = distance(line1.end.world_position, line2.end.world_position)
            if dist < .1:
                # print('YAY!', dist)
                break
            else:
                line1.rotation_x -= 1
                line2.rotation_x -= 1

        trail = WhiteCube()
        trail.scale *= .05
        trail.origin_z = -.5
        trail.world_position = line1.world_position + (line1.forward * .5)
        destroy(trail, 1)

    intersecting_point.update = intersect_update

    slider = Slider(0, 4, default=1, dynamic=True, y=-.4)
    slider.label.text = 'cube height'
    def on_value_changed():
        line2.scale_z = slider.value
    slider.on_value_changed = on_value_changed

    class Motor(Entity):
        def __init__(self, motor):
            self.motor = motor
            super().__init__()

        def update(self):
            self.motor.rotation_z += held_keys['d'] * 2
            self.motor.rotation_z -= held_keys['a'] * 2

    Motor(line0)

    EditorCamera()
    camera.orthographic = True
    camera.fov = 10
    app.run()



from ursina import *

# Main Menu Example, or it can be any kind of menu, like Inventory, Quest journal, etc.
# Created by Doctor
# 09 Feb 21

# Edited version by @jasonheller on 07 Apr 22

# Class of game menu
class MenuMenu(Entity):
    def __init__(self, **kwargs):
        super().__init__(parent=camera.ui, ignore_paused=True)

        # Create empty entities that will be parents of our menus content
        self.main_menu = Entity(parent=self, enabled=True)
        self.options_menu = Entity(parent=self, enabled=False)
        self.help_menu = Entity(parent=self, enabled=False)

        # Add a background. You can change 'shore' to a different texture of you'd like.
        self.background = Sprite('shore', color=color.dark_gray, z=1)

        # [MAIN MENU] WINDOW START
        # Title of our menu
        Text("MAIN MENU", parent=self.main_menu, y=0.4, x=0, origin=(0,0))

        def switch(menu1, menu2):
            menu1.enable()
            menu2.disable()

        # Button list
        ButtonList(button_dict={
            "Start": Func(print_on_screen,"You clicked on Start button!", position=(0,.1), origin=(0,0)),
            "Options": Func(lambda: switch(self.options_menu, self.main_menu)),
            "Help": Func(lambda: switch(self.help_menu, self.main_menu)),
            "Exit": Func(lambda: application.quit())
        },y=0,parent=self.main_menu)
        # [MAIN MENU] WINDOW END

        # [OPTIONS MENU] WINDOW START
        # Title of our menu
        Text ("OPTIONS MENU", parent=self.options_menu, y=0.4, x=0, origin=(0, 0))

        # Button
        Button("Back",parent=self.options_menu,y=-0.3,scale=(0.1,0.05),color=rgb(50,50,50),
               on_click=lambda: switch(self.main_menu, self.options_menu))

        # [OPTIONS MENU] WINDOW END

        # [HELP MENU] WINDOW START
        # Title of our menu
        Text ("HELP MENU", parent=self.help_menu, y=0.4, x=0, origin=(0, 0))

        # Button list
        ButtonList (button_dict={
            "Gameplay": Func(print_on_screen,"You clicked on Gameplay help button!", position=(0,.1), origin=(0,0)),
            "Battle": Func(print_on_screen,"You clicked on Battle help button!", position=(0,.1), origin=(0,0)),
            "Control": Func(print_on_screen,"You clicked on Control help button!", position=(0,.1), origin=(0,0)),
            "Back": Func (lambda: switch(self.main_menu, self.help_menu))
        }, y=0, parent=self.help_menu)
        # [HELP MENU] WINDOW END

        # Here we can change attributes of this class when call this class
        for key, value in kwargs.items ():
            setattr (self, key, value)

    # Input function that check if key pressed on keyboard
    def input(self, key):
        # And if you want use same keys on different windows
        # Like [Escape] or [Enter] or [Arrows]
        # Just write like that:

        # If our main menu enabled and we press [Escape]
        if self.main_menu.enabled and key == "escape":
                application.quit()
        elif self.options_menu.enabled and key == "escape":
            self.main_menu.enable()
            self.options_menu.disable()
        elif self.help_menu.enabled and key == "escape":
            self.main_menu.enable()
            self.help_menu.disable()

    # Update function that check something every frame
    # You can use it similar to input with checking
    # what menu is currently enabled
    def update(self):
        pass


# Init application
app = Ursina(title='Main Menu Tutorial')

# Call our menu
main_menu = MenuMenu()

# Run application
app.run()



from ursina import *

# window.set_z_order(window.Z_top)
app = Ursina()
# window.position=(1920/2, 0)
# window.size = (1920/2, 1080/2)

# ok guys, let's start by creating a ground, uwu
# ground = Entity()
# and set the model by giving it the name of our model.
#let's just use the included 'plane' model for now.
# ground = Entity(model='plane')
# scale it up and give it a color.
class Test(Entity):
    def __init__(self, **kwargs):
        super().__init__()
        self.ground = Entity(model='plane', scale=10, color=color.green)
        self.player = Entity(model='cube', scale_y=2, origin_y=-.5, color=color.azure)

        for key, value in kwargs.items():
            setattr(self, key, value)
# set the camera position so we can see better.
# camera.position = (1,10,-20)
# camera.look_at(ground)
# EditorCamera(eternal = True)
# camera.add_script(SmoothFollow(target=player, offset=(0,3,-20)))

    def update(self):
        self.player.x += held_keys['d'] * time.dt
        self.player.x -= held_keys['a'] * time.dt
        #
        # if t > 1:
        #     application.hot_reloader.reload_code()
        #     t = 0

Test()

app.run()



from ursina import Shader


camera_grayscale_shader = Shader(
fragment='''
#version 430

uniform sampler2D tex;
in vec2 uv;
out vec4 color;

void main() {
    color = texture(tex, uv).rgba;
    float water_level = 0;

    if (color == vec4(1,0,0,1)) {
        vec2 new_uv = vec2(uv.x, -uv.y + -.365);
        color = texture(tex, new_uv).rgba * 0.75;
    }
    // color = vec4(1,1,0,1);
    // if (uv.y < .25) {
    //     color = vec4(0,0.,1.,1.);
    // }
    // color = vec4(rgb, 1.0);
}

''')



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    # e = Entity(model='sphere', color=color.orange)
    # e = Entity(model='cube', y=-1)
    camera.shader = camera_grayscale_shader
    # camera.clip_plane_near = 1
    EditorCamera()

    # random.seed(2)
    # for i in range(20):
    #     e = Entity(model='cube', position=Vec3(random.random(),random.random(),random.random())*2, rotation=Vec3(random.random(),random.random(),random.random())*360)
    #
    plane = Entity(model='quad', scale=(20, 3), y=-3, color=color.red)
    Sprite('shore', z=2)
    def update():
        camera.x += (held_keys['d'] - held_keys['a']) * time.dt * 2
    app.run()



# from OpenGL.GLUT import *
# from OpenGL.GLU import *
# from OpenGL.GL import *
# import sys
#
# name = 'ball_glut'
#
# def main():
#     glutInit(sys.argv)
#     glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
#     glutInitWindowSize(400,400)
#     glutCreateWindow(name)
#
#     glClearColor(0.,0.,0.,1.)
#     glShadeModel(GL_SMOOTH)
#     glEnable(GL_CULL_FACE)
#     glEnable(GL_DEPTH_TEST)
#     glEnable(GL_LIGHTING)
#     lightZeroPosition = [10.,4.,10.,1.]
#     lightZeroColor = [0.8,1.0,0.8,1.0] #green tinged
#     glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition)
#     glLightfv(GL_LIGHT0, GL_DIFFUSE, lightZeroColor)
#     glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1)
#     glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05)
#     glEnable(GL_LIGHT0)
#     glutDisplayFunc(display)
#     glMatrixMode(GL_PROJECTION)
#     gluPerspective(40.,1.,1.,40.)
#     glMatrixMode(GL_MODELVIEW)
#     gluLookAt(0,0,10,
#               0,0,0,
#               0,1,0)
#     glPushMatrix()
#     glutMainLoop()
#     return
#
# def display():
#     glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT)
#     glPushMatrix()
#     color = [1.0,0.,0.,1.]
#     glMaterialfv(GL_FRONT,GL_DIFFUSE,color)
#     glutSolidSphere(2,20,20)
#     glPopMatrix()
#     glutSwapBuffers()
#     return
#
# if __name__ == '__main__':
#     main()

#
    # An example of a Win32 window and an OpenGL context for rendering graphics
    #
    # Visit https://devs.site for more
    #

# test.py
from time import perf_counter
t = perf_counter()
# import moderngl_window as mglw
from moderngl_window import WindowConfig, run_window_config

class Test(WindowConfig):
    gl_version = (3, 3)

    def render(self, time, frametime):
        self.ctx.clear(1.0, 0.0, 0.0, 0.0)

print('------------', perf_counter() - t)
# run_window_config(Test)
run_window_config(Test, '--window tk')



from ursina import *

class Player(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'play'
        self.model = 'quad'

        # adds a Weapon instance to self.scripts and andds self.weapon
        # self gets added to weapon instance as 'entity'
        # self.add_script(Weapon()) # instance
        # self.add_script('weapon') # module name
        # self.add_script('Weapon') # class name
        # printvar(self.weapon)
        # self.weapon = Weapon()
        # print(self.weapon)

    def attack(self):
        print('ATTACK!')


    def input(self, key):
        if key == 'd':
            destroy(self.weapon)

class Weapon():
    def __init__(self):
        self.damage = 10

    def input(self, key):
        if key == 'e':
            self.entity.animate_color(color.yellow, duration=.5)
            self.entity.animate_color(color.white, delay=2, duration=1.5)
            self.entity.attack()

class Tester(Entity):
    def __init__(self):
        super().__init__()

    def input(self, key):
        if key == 't':
            print('t')
            pass
            # print(len(globals()))
            # print(app.player.weapon)
            # for g in globals():
            #     print(g)
                # try:
                #     print(g.name)
                # except:
                #     pass


app = Ursina()
t = Tester()
app.player = Player()
app.run()



from pathlib import Path
from ursina import application
from time import perf_counter
from ursina.string_utilities import print_info


def ursinamesh_to_obj(mesh, name='', out_path:Path=None, max_decimals=5, flip_faces=True):
    from ursina.string_utilities import camel_to_snake
    from ursina.array_tools import chunk_list


    if not name:
        name = camel_to_snake(mesh.__class__.__name__)

    obj = ''
    obj += f'mtllib {name}.mtl\n'
    obj += f'usemtl {name}\n'
    obj += f'o {name}\n'

    # Vertices
    for v in mesh.vertices:
        obj += f'v {round(v[0], max_decimals)} {round(v[1], max_decimals)} {round(v[2], max_decimals)}\n'

    # UVs
    has_uvs = bool(mesh.uvs)
    if has_uvs:
        for uv in mesh.uvs:
            obj += f'vt {round(uv[0], max_decimals)} {round(uv[1], max_decimals)}\n'

    # Normals
    has_normals = bool(mesh.normals)
    if has_normals:
        for n in mesh.normals:
            obj += f'vn {round(n[0], max_decimals)} {round(n[1], max_decimals)} {round(n[2], max_decimals)}\n'

    obj += 's off\n'

    # Triangles
    tris = mesh.triangles if mesh.triangles else chunk_list(mesh.indices, 3)

    for tri in tris:
        if flip_faces:
            tri = (tri[2], tri[1], tri[0])  # reverse winding

        obj += 'f'
        for idx in tri:
            v_idx = idx + 1  # OBJ is 1-indexed
            vt_idx = v_idx if has_uvs else ''
            vn_idx = v_idx if has_normals else ''

            if has_uvs and has_normals:
                obj += f' {v_idx}/{vt_idx}/{vn_idx}'
            elif has_uvs:
                obj += f' {v_idx}/{vt_idx}'
            elif has_normals:
                obj += f' {v_idx}//{vn_idx}'
            else:
                obj += f' {v_idx}'
        obj += '\n'

    out_file = out_path / f'{name}.obj'
    out_file.parent.mkdir(parents=True, exist_ok=True)
    with open(out_file, 'w') as f:
        f.write(obj)

    print_info(f'saved obj: {out_file}')




def ursinamesh_to_dae(mesh, name, folder:Path=application.models_compressed_folder, texture_name=''):
    num_vertices = len(mesh.generated_vertices)
    vertices = ' '.join([f'{v[2]} {v[1]} {v[0]}' for v in mesh.generated_vertices])

    # triangle_indices = ' '.join([f'{i} '*4 for i in range(num_vertices, 0, -1)])
    triangle_indices = ' '.join([f'{i} '*4 for i in range(num_vertices)])
    num_triangle_indices = num_vertices

    num_uvs = len(mesh.uvs)
    uvs = ' '.join([f'{uv[0]} {uv[1]}' for uv in mesh.uvs])

    num_vertex_colors = len(mesh.colors)
    vertex_colors = ' '.join([f'{c[0]} {c[1]} {c[2]} {c[3]}' for c in mesh.colors])

    texture_name = texture_name.replace('.','_')
    # print(vertices)
    text = f'''<?xml version="1.0" encoding="utf-8"?>
<COLLADA xmlns="http://www.collada.org/2005/11/COLLADASchema" version="1.4.1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <asset>
    <contributor>
      <author>Ursina User</author>
      <authoring_tool>ursina</authoring_tool>
    </contributor>
    <unit name="meter" meter="1"/>
    <up_axis>Y_UP</up_axis>
  </asset>
  <library_effects>
    <effect id="Material-effect">
      <profile_COMMON>
        <newparam sid="{texture_name}-surface">
          <surface type="2D">
            <init_from>{texture_name}</init_from>
          </surface>
        </newparam>
        <newparam sid="{texture_name}-sampler">
          <sampler2D>
            <source>{texture_name}-surface</source>
          </sampler2D>
        </newparam>
        <technique sid="common">
          <lambert>
            <emission>
              <color sid="emission">0 0 0 1</color>
            </emission>
            <diffuse>
              <texture texture="{texture_name}-sampler" texcoord="UVMap"/>
            </diffuse>
            <index_of_refraction>
              <float sid="ior">1.45</float>
            </index_of_refraction>
          </lambert>
        </technique>
      </profile_COMMON>
    </effect>
  </library_effects>
  <library_images/>
  <library_materials>
    <material id="Material-material" name="Material">
      <instance_effect url="#Material-effect"/>
    </material>
  </library_materials>
  <library_geometries>
    <geometry id="Cube-mesh" name="Cube">
      <mesh>
        <source id="Cube-mesh-positions">
          <float_array id="Cube-mesh-positions-array" count="{num_vertices*3}">{vertices}</float_array>
          <technique_common>
            <accessor source="#Cube-mesh-positions-array" count="{num_vertices}" stride="3">
              <param name="X" type="float"/>
              <param name="Y" type="float"/>
              <param name="Z" type="float"/>
            </accessor>
          </technique_common>
        </source>
        <source id="Cube-mesh-normals">
          <float_array id="Cube-mesh-normals-array" count="18">0 0 1 0 -1 0 -1 0 0 0 0 -1 1 0 0 0 1 0</float_array>
          <technique_common>
            <accessor source="#Cube-mesh-normals-array" count="6" stride="3">
              <param name="X" type="float"/>
              <param name="Y" type="float"/>
              <param name="Z" type="float"/>
            </accessor>
          </technique_common>
        </source>
        <source id="Cube-mesh-map-0">
          <float_array id="Cube-mesh-map-0-array" count="{num_uvs*2}">{uvs}</float_array>
          <technique_common>
            <accessor source="#Cube-mesh-map-0-array" count="{num_uvs}" stride="2">
              <param name="S" type="float"/>
              <param name="T" type="float"/>
            </accessor>
          </technique_common>
        </source>
        <source id="Cube-mesh-colors-Col" name="Col">
          <float_array id="Cube-mesh-colors-Col-array" count="{num_vertex_colors*3}">{vertex_colors}</float_array>
          <technique_common>
            <accessor source="#Cube-mesh-colors-Col-array" count="{num_vertex_colors}" stride="4">
              <param name="R" type="float"/>
              <param name="G" type="float"/>
              <param name="B" type="float"/>
              <param name="A" type="float"/>
            </accessor>
          </technique_common>
        </source>
        <vertices id="Cube-mesh-vertices">
          <input semantic="POSITION" source="#Cube-mesh-positions"/>
        </vertices>
        <triangles material="Material-material" count="{num_triangle_indices}">
          <input semantic="VERTEX" source="#Cube-mesh-vertices" offset="0"/>
          <input semantic="NORMAL" source="#Cube-mesh-normals" offset="1"/>
          <input semantic="TEXCOORD" source="#Cube-mesh-map-0" offset="2" set="0"/>
          <input semantic="COLOR" source="#Cube-mesh-colors-Col" offset="3" set="0"/>
          <p>{triangle_indices}</p>
        </triangles>
      </mesh>
    </geometry>
  </library_geometries>
  <library_visual_scenes>
    <visual_scene id="Scene" name="Scene">
      <node id="Cube" name="Cube" type="NODE">
        <matrix sid="transform">1 0 0 0 0 1 0 0 0 0 1 0 0 0 0 1</matrix>
        <instance_geometry url="#Cube-mesh" name="Cube">
          <bind_material>
            <technique_common>
              <instance_material symbol="Material-material" target="#Material-material">
                <bind_vertex_input semantic="UVMap" input_semantic="TEXCOORD" input_set="0"/>
              </instance_material>
            </technique_common>
          </bind_material>
        </instance_geometry>
      </node>
    </visual_scene>
  </library_visual_scenes>
  <scene>
    <instance_visual_scene url="#Scene"/>
  </scene>
</COLLADA>
    '''
    with (folder / f'{name}.dae').open('w') as f:
        f.write(text)


if __name__ == '__main__':
    from ursina import Ursina, Entity, load_model, EditorCamera, Sky
    app = Ursina()

    t = perf_counter()
    Entity(model='untitled')
    print('-------', perf_counter() - t)
    m = load_model('cube', use_deepcopy=True)
    ursinamesh_to_dae(m, 'dae_export_test.dae')
    EditorCamera()
    Sky(texture='sky_sunset')

    app.run()



from copy import deepcopy

from ursina import *


class Quad(Mesh):
    _cache = {}
    _corner_maker = None
    _point_placer = None

    def __new__(cls, radius=.1, segments=8, aspect=1, scale=(1,1), mode='ngon', thickness=1):
        # special case: plain quad
        if radius == 0 and aspect == 1 and scale == (1, 1) and mode == 'ngon':
            return Mesh(
                vertices=[
                    Vec3(-0.5, -0.5, 0.0), Vec3(0.5, -0.5, 0.0),
                    Vec3(0.5, 0.5, 0.0), Vec3(-0.5, 0.5, 0.0)
                ],
                triangles=[(0, 1, 2, 3)],
                uvs=[Vec2(0, 0), Vec2(1, 0), Vec2(1, 1), Vec2(0, 1)],
                mode='triangle'
            )

        key = (radius, segments, aspect, scale, mode, thickness)
        if key in cls._cache:
            try:
                return deepcopy(cls._cache[key])
            except:     # deepcopy can fail if the model has been destroyed
                pass
        instance = super().__new__(cls)
        cls._cache[key] = instance
        return instance

    def __init__(self, radius=.1, segments=8, aspect=1, scale=(1,1), mode='ngon', thickness=1):
        if hasattr(self, '_initialized'):
            return
        self._initialized = True

        if not Quad._corner_maker:
            Quad._corner_maker = Entity(eternal=True, add_to_scene_entities=False)
        if not Quad._point_placer:
            Quad._point_placer = Entity(parent=Quad._corner_maker, x=-radius, eternal=True, add_to_scene_entities=False)

        super().__init__()
        self.radius = radius
        self.mode = mode
        self.thickness = thickness

        self.vertices = [Vec3(0, 0, 0), Vec3(1, 0, 0), Vec3(1, 1, 0), Vec3(0, 1, 0)]

        _segments = segments + 1
        if _segments > 1:
            new_verts = []
            Quad._corner_maker.rotation_z = -90 / _segments / 2
            Quad._corner_maker.position = Vec3(0, 0, 0)
            Quad._point_placer.position = Vec3(-radius, 0, 0)

            corrections = [Vec3(radius, radius, 0), Vec3(-radius, radius, 0),
                           Vec3(-radius, -radius, 0), Vec3(radius, -radius, 0)]

            for j in range(4):
                Quad._corner_maker.position = self.vertices[j] + corrections[j]
                for i in range(_segments):
                    new_verts.append(Quad._point_placer.world_position)
                    Quad._corner_maker.rotation_z -= 90 / _segments

            self.vertices = new_verts

        self.uvs = [Vec2(v[0], v[1]) for v in self.vertices]

        # aspect correction
        for v in self.vertices:
            if v[0] < .5:
                v[0] /= aspect
            else:
                v[0] = lerp(v[0], 1, 1 - (1 / aspect))

        # scale correction
        for v in self.vertices:
            if v[0] > .5:
                v[0] += (scale[0] - 1)
            if v[1] > .5:
                v[1] += (scale[1] - 1)

        # center mesh
        offset = sum(self.vertices) / len(self.vertices)
        self.vertices = [(v[0] - offset[0], v[1] - offset[1], v[2] - offset[2]) for v in self.vertices]

        if mode == 'line':
            self.vertices.append(self.vertices[0])

        self.normals = [Vec3.back for _ in self.vertices]
        self.generate()


if __name__ == '__main__':
    app = Ursina()
    from time import perf_counter
    t = perf_counter()
    # m =
    for i in range(100):
        Entity(model=Quad(scale=(3,1), thickness=3, segments=3, mode='line'), color = color.hsv(0,1,1,.7))
    # Entity(scale=(3,1), model=Quad(aspect=3), color = color.hsv(60,1,1,.3))
    print('-------', (perf_counter() - t))

    origin = Entity(model='quad', color=color.orange, scale=(.05, .05))
    # ed = EditorCamera(rotation_speed = 200, panning_speed=200)

    Entity(model=Quad(0), texture='shore', x=-1)

    camera.z = -5
    app.run()



from panda3d.core import Shader


shader_code = '''

#define SAMPLES 8
#define PATHDEPTH 2

#define RAYCASTSTEPS 1
#define RAYCASTSTEPSRECURSIVE 2

#define EPSILON 0.001
#define MAXDISTANCE 180. //180
#define MAXHEIGHT 30.
#define SPEED 1.0
float time;
vec3 normal;

void vshader(float4 vtx_position : POSITION,
            float2 vtx_texcoord0 : TEXCOORD0,
            out float4 l_position : POSITION,
            out float2 l_texcoord0 : TEXCOORD0,
            uniform float4 texpad_tex,
            uniform float4x4 mat_modelproj)
{
    l_position=mul(mat_modelproj, vtx_position);
    l_texcoord0 = vtx_position.xy * texpad_tex.xy + texpad_tex.xy;
}

half3 AdjustContrast(half3 color, half contrast) {
    return saturate(lerp(half3(0.5, 0.5, 0.5), color, contrast));
}
float hash( const float n ) {
	return fract(sin(n)*43758.54554213);
}
float2 hash2( const float n ) {
	return fract(sin(float2(n,n+1.))*float2(43758.5453123));
}
float2 hash2( const float2 n ) {
	return fract(sin(float2( n.x*n.y, n.x+n.y))*float2(25.1459123,312.3490423));
}
float3 hash3( const float2 n ) {
	return fract(sin(float3(n.x, n.y, n+2.0))*float3(36.5453123,43.1459123,11234.3490423));
}

float intersectPlane( const float3 cam_pos, const float3 rd, const float height) {
	if (rd.y==0.0) return 500.;
	float d = -(cam_pos.y - height)/rd.y;
	if( d > 0. ) {
		return d;
	}
	return 500.;
}

float intersectUnitSphere ( const float3 cam_pos, const float3 rd, const float radius, const float3 sph ) {
	float3  ds = cam_pos - sph;
	float b = dot( rd, ds );
	float cs = dot( ds, ds ) - (radius*radius);
	float discriminant = b*b - cs;

	if( discriminant > 0.0 ) {
		float t = -b - sqrt( discriminant );
        // float t = (-b - sqrt(discriminant)) / (2.0 * a);
		if( t > 0. ) {
			return t;
		}
	}
	return 500.;
}


//
// Scene
//

// void getSphereOffset( const float2 grid, out float2 center ) {
// 	center = (hash2( grid+float2(43.12,1.23) ) - float2(0.5) )*(GRIDSIZESMALL);
//     center = 0;
// }
// void getSpherePosition( const float2 grid, const float2 sphereOffset, out float3 center ) {
// 	float2 offset = grid + sphereOffset;
// 	center = float3( offset.x, 0., offset.y ) + 0.5*float3( GRIDSIZE, 2., GRIDSIZE );
// }
// float3 getSphereColor( const float2 grid ) {
// 	//float3 col = hash3( grid+float2(43.12*grid.y,12.23*grid.x) );
//     float3 col = (1,1,1);
//     return mix(col,col*col,.8);
// }

float3 getBackgroundColor( const float3 cam_pos, const float3 rd ) {
	return 1.4*mix(float3(.5),float3(.6,.8,.9), .5+.5*rd.y);
}

float3 trace(const float3 cam_pos, const float3 rd, out float3 intersection, out float3 normal,
           out float dist, out int material, const int steps,
           TRACE_FUNC_INPUT
       ) {
	dist = MAXDISTANCE;
	float distcheck;

	float3 col, normalcheck;

	material = 0;
	col = getBackgroundColor(cam_pos, rd);

    // ground plane
	if( (distcheck = intersectPlane( cam_pos,  rd, 0.)) < MAXDISTANCE ) {
		dist = distcheck;
		material = 1;
		normal = float3(0, 1, 0);
		col = float3(.2, .5, .05);
	}

	// trace grid

	float3 pos = cam_pos;
	float3 ri = 1.0/rd;
	float3 rs = sign(rd);
	float3 dis = (pos-cam_pos + 0.5  + rs*0.5) * ri;
	float3 mm = float3(0.0);
	float2 offset;

    float radius = 4;

	for( int i=0; i<steps; i++ )	{
		if( material == 2 || distance( cam_pos.xz, pos.xz ) > dist ) break; {
            DISTCHECK_CODE


			// mm = step(dis.xyz, dis.zyx);
			// dis += mm * rs * ri;
			// pos += mm * rs;
		}
	}

	intersection = cam_pos+rd*dist;

	return col;
}

float2 rv2;

float3 cosWeightedRandomHemisphereDirection2( const float3 n ) {
	float3  uu = normalize( cross( n, float3(0.0,1.0,1.0) ) );
	float3  vv = cross( uu, n );

	float ra = sqrt(rv2.y);
	float rx = ra*cos(6.2831*rv2.x);
	float ry = ra*sin(6.2831*rv2.x);
	float rz = sqrt( 1.0-rv2.y );
	float3  rr = float3( rx*uu + ry*vv + rz*n );

    return normalize( rr );
}

void fshader(float2 l_texcoord0 : TEXCOORD0,
             out float4 o_color : COLOR,
             uniform sampler2D k_tex : TEXUNIT0,
             uniform float2 iResolution,
             uniform float k_iTime,
	           uniform float k_iFrame,
             uniform float3 k_camera_rotation,
             uniform float3 k_camera_forward,
             uniform float3 k_camera_position,
             FSHADER_INPUT
             // uniform float3 k_sphere_0_position,
             // uniform float3 k_sphere_1_position,
             // uniform float k_sphere_0_radius,
             // uniform float k_sphere_1_radius
         )
{
    float2 uv = l_texcoord0/iResolution.xy;
    // float2 uv = l_texcoord0;
    // float3 col = 0.5 + 0.5*cos(k_iTime+uv.xyx+float3(0,.4,.8));

    time = k_iTime;
        // float2 q = fragCoord.xy/iResolution.xy;
    float2 q = l_texcoord0.xy;
  	float2 p = -1.0+2.0*q;
  	p.x *= iResolution.x/iResolution.y;

  	float3 col = float3( 0. );

  	// raytrace
  	int material;
  	float3 normal, intersection;
  	float dist;
  	float seed = time+(p.x+iResolution.x*p.y)*1.51269341231;

  	for( int j=0; j<SAMPLES + min(0,k_iFrame); j++ ) {
  		float fj = float(j);


  		rv2 = hash2( 24.4316544311*fj+time+seed );

  		float2 pt = p+rv2/(0.5*iResolution.xy);

  		// camera
  		// float3 cam_pos = float3( cos( 0.232*time) * 10., 6.+3.*cos(0.3*time), GRIDSIZE*(time/SPEED) );
  		// float3 look_dir = cam_pos + float3( -sin( 0.232*time) * 10., -2.0+cos(0.23*time), 10.0 );
  		float3 cam_pos = k_camera_position; // position
  		float3 look_dir = cam_pos + k_camera_forward;

        // look_dir = float3(k_camera_rotation.y, -2.0+cos(0.23), 10)
  		// float roll = -0.15*sin(0.5*time);
        float roll = 0;
  		// camera tx
  		float3 cw = normalize( look_dir-cam_pos );
  		float3 cp = float3( sin(roll), cos(roll), 0.0 );
  		float3 cu = normalize( cross(cw,cp) );
  		float3 cv = normalize( cross(cu,cw) );

  		float3 rd = normalize( pt.x*cu + pt.y*cv + 1.5*cw );

  		float3 colsample = float3( 1. );

  		// first hit
  		rv2 = hash2( (rv2.x*2.4543263+rv2.y)*(time+1.) );
  		colsample *= trace(cam_pos, rd, intersection, normal, dist, material, RAYCASTSTEPS,
                TRACE_CALL_CODE
            );

  		// bounces
  		for( int i=0; i<(PATHDEPTH-1); i++ ) {
  			if( material != 0 ) {
  				rd = cosWeightedRandomHemisphereDirection2( normal );
  				cam_pos = intersection + EPSILON*rd;

  				rv2 = hash2( (rv2.x*2.4543263+rv2.y)*(time+1.)+(float(i+1)*.23) );

  				colsample *= trace(cam_pos, rd, intersection, normal, dist, material, RAYCASTSTEPSRECURSIVE,
                    TRACE_CALL_CODE
                        // k_sphere_0_position, k_sphere_0_radius,
                        // k_sphere_1_position, k_sphere_1_radius
                    );
  			}
  		}
  		colsample = sqrt(clamp(colsample, 0., 1.));
  		if( material == 0 ) {
  			col += colsample;
  		}
  	}
  	col  /= float(SAMPLES);




    o_color = float4(col, 1.0);
}


'''



if __name__ == '__main__':
    from ursina import *
    from PIL import Image
    # from panda3d.core import Texture as PandaTexture
    window.vsync = False
    app = Ursina()

    class Player(Entity):
        def __init__(self, **kwargs):
            super().__init__(model='sphere', color=color.red, **kwargs)
            self.speed = 10

        def update(self):
            h = max((held_keys['gamepad left stick x'], held_keys['d']-held_keys['a']), key=lambda x: abs(x))
            v = max((held_keys['gamepad left stick y'], held_keys['w']-held_keys['s']), key=lambda x: abs(x))
            self.direction = Vec3(h,0,v).normalized()
            self.input_strength = min(Vec3(h,0,v).length(), 1)
            self.velocity = self.direction * time.dt * self.speed * self.input_strength
            self.position += self.velocity

    Player()

    quad = Entity(model='quad', texture=Texture(Image.new(mode="RGBA", size=(1920//8,1080//8))), scale=(16,9), always_on_top=True)
    quad.world_parent=camera
    # quad = Entity(parent=camera.ui, model='quad', texture=Texture(Image.new(mode="RGBA", size=(1920,1080))), scale=(camera.aspect_ratio, 1))
    quad.texture.filtering = None

    camera.position = (0,0,0)
    camera_rotation = [0,0,0]
    dummy_object = camera
    direction = Entity(parent=dummy_object, z=1)
    t = 0
    frame = 0

    random.seed(0)
    spheres = list()
    for z in range(3):
        for y in range(3):
            for x in range(3):
                Entity(model='sphere', position=(x*5,y*5,z*5), color=color.random_color(), scale=random.uniform(.1,10))

    spheres = [e for e in scene.entities if e.model and e.model.name == 'sphere']
    # for e in scene.entities:
    #     if e.model:
    #         print(e.name, e.model.name)

    trace_func_input = '\n'.join([f'const float3 sphere_{i}_position, const float sphere_{i}_radius, ' for i in range(len(spheres))])[:-2]
    fshader_input = '\n'.join([f'uniform float3 k_sphere_{i}_position, uniform float k_sphere_{i}_radius, ' for i in range(len(spheres))])[:-2]
    trace_call_code = '\n'.join([f'k_sphere_{i}_position, k_sphere_{i}_radius, ' for i in range(len(spheres))])[:-2]

    distcheck_code = ''
    for i, sphere in enumerate(spheres):
        distcheck_code += f'''
        if( (distcheck = intersectUnitSphere( cam_pos, rd, sphere_{i}_radius, sphere_{i}_position )) < dist ) {{
            dist = distcheck;
            normal = normalize((cam_pos+rd*dist)-sphere_{i}_position);
            col = float3({sphere.color.r},{sphere.color.g},{sphere.color.b});
            material = 2;
        }}'''



    # print('----------', fshader_input)
    shader_code = shader_code.replace('TRACE_FUNC_INPUT', trace_func_input)
    shader_code = shader_code.replace('DISTCHECK_CODE', distcheck_code)
    shader_code = shader_code.replace('FSHADER_INPUT', fshader_input)
    shader_code = shader_code.replace('TRACE_CALL_CODE', trace_call_code)
    # for i, l in enumerate(shader_code.split('\n')):
    #     print(i, l)
    from panda3d.core import Shader
    camera_contrast_shader = Shader.make(shader_code, Shader.SL_Cg)
    quad.shader = camera_contrast_shader
    quad.set_shader_input('iTime', 0)
    quad.set_shader_input('iFrame', 0)
    # quad.set_shader_input('iResolution', window.size / 2)
    quad.set_shader_input('iResolution', quad.texture.size)
    quad.set_shader_input('tex', quad.texture._texture)
    quad.set_shader_input('camera_rotation', camera.position)
    quad.set_shader_input('camera_forward', camera.right)
    quad.set_shader_input('camera_position', (0,0,0))

    for i, e in enumerate(spheres):
        # print(i, e.model.name)
        quad.set_shader_input(f'sphere_{i}_position', (-e.world_x*2, e.world_y*2, e.world_z*2))
        quad.set_shader_input(f'sphere_{i}_radius', e.world_scale[0])


    def update():
        global t, frame, camera_rotation, spheres
        t += time.dt

        quad.set_shader_input('iTime', t)
        quad.set_shader_input('iFrame', frame)
        dir = direction.world_position-dummy_object.world_position
        dir[0] *= -1
        quad.set_shader_input('camera_forward', dir)
        quad.set_shader_input('camera_position', (-dummy_object.world_x, dummy_object.world_y, dummy_object.world_z))

        for i, e in enumerate(spheres):
            quad.set_shader_input(f'sphere_{i}_position', (-e.world_x*2, e.world_y*2, e.world_z*2))
            quad.set_shader_input(f'sphere_{i}_radius', e.world_scale[0])



    def input(key):
        if key == 'tab':
            quad.enabled = not quad.enabled


    camera.position=(0,1,-10)
    EditorCamera()
    window.fps_counter.color = color.black
    # window.size /= 2
    window.size = (400,300)
    app.run()



# how to move rotate the
# camera around a point.

from ursina import *
from ursina.shaders import lit_with_shadows_shader #!hide
Entity.default_shader = lit_with_shadows_shader #!hide

app = Ursina()
window.size = Vec2(1080, 1920) * .5 #!hide
window.position = (0,0) #!hide

# Make an empty Entity to help
# us rotate.
origin = Entity()

# Parent the camera to the entity.
camera.parent = origin

# Move the camera back.
camera.z = -20

# Rotate when we right click and
# drag.
def update():
  origin.rotation_y += (
    # How far the mouse moved
    # horizontally since last
    # update.
    mouse.velocity[0]
    # Only rotate if right
    # mouse button is being
    # held.
    * mouse.right * 100
    )
  # Do the same vertically.
  origin.rotation_x -= (
    mouse.velocity[1]
    * mouse.right * 100
    )

# Create a cube on some grass
# so we have something to
# look at :D
player = Entity(
  model='cube',
  origin_y=-.5,
  color=color.orange
  )
grass = Entity(
  model='plane',
  scale=40,
  texture='grass',
  texture_scale=Vec2(10)
  )
DirectionalLight().look_at(Vec3(.5,-1,.5)) #!hide
Sky() #!hide
# Run program!
app.run()



from ursina import *
from ursina.networking import *

app = Ursina(borderless=False)

start_text = "Press H to host or C to connect."
status_text = Text(text=start_text, origin=(0, 0), z=1)

box = Entity(model="quad", texture="white_cube", color=color.red, parent=camera.ui, scale=0.1, y=0.1)
other_box = Entity(model="quad", texture="white_cube", color=color.blue, parent=camera.ui, scale=0.1, y=-0.1, z=0.5)
other_box.new_x = other_box.x
other_box.new_y = other_box.y
other_box.prev_x = other_box.x
other_box.prev_y = other_box.y

box_speed = 0.4

update_rate = 1.0 / 20.0
update_timer = 0.0

lerp_time = update_rate * 1.25
lerp_timer = 0.0

peer = RPCPeer()

@rpc(peer)
def set_position(connection, time_received, position: Vec2):
    global lerp_timer

    other_box.x = other_box.new_x
    other_box.y = other_box.new_y
    other_box.prev_x = other_box.x
    other_box.prev_y = other_box.y
    other_box.new_x = position.x
    other_box.new_y = position.y

    lerp_timer = 0.0

def update():
    global update_timer, lerp_timer 

    peer.update()
    if not peer.is_running():
        status_text.text = start_text
        box.x = 0.0
        box.y = 0.1
        other_box.x = 0.0
        other_box.y = -0.1
        return

    if peer.is_hosting():
        status_text.text = "Hosting on localhost, port 8080.\nWASD to move."
    else:
        status_text.text = "Connected to host with address localhost, port 8080.\nWASD to move."

    box.x += (held_keys["d"] - held_keys["a"]) * box_speed * time.dt
    box.y += (held_keys["w"] - held_keys["s"]) * box_speed * time.dt

    lerp_timer += time.dt
    other_box.x = lerp(other_box.prev_x, other_box.new_x, lerp_timer / lerp_time)
    other_box.y = lerp(other_box.prev_y, other_box.new_y, lerp_timer / lerp_time)
    if lerp_timer >= lerp_time:
        other_box.x = other_box.new_x
        other_box.y = other_box.new_y
        other_box.prev_x = other_box.x
        other_box.prev_y = other_box.y
        other_box.new_x = other_box.x
        other_box.new_y = other_box.y
        lerp_timer = 0.0

    update_timer += time.dt
    if update_timer >= update_rate:
        if peer.is_running() and peer.connection_count() > 0:
            peer.set_position(peer.get_connections()[0], Vec2(box.x, box.y))
        update_timer = 0.0

def input(key):
    if key == "h":
        box.y = 0.1
        other_box.y = -0.1
        other_box.new_x = other_box.x
        other_box.new_y = other_box.y
        other_box.prev_x = other_box.x
        other_box.prev_y = other_box.y

        peer.start("localhost", 8080, is_host=True)
    elif key == "c":
        box.y = -0.1
        other_box.y = 0.1
        other_box.new_x = other_box.x
        other_box.new_y = other_box.y
        other_box.prev_x = other_box.x
        other_box.prev_y = other_box.y

        peer.start("localhost", 8080, is_host=False)

app.run()



from ursina import Shader


curvature_shader = Shader(
fragment='''
#version 430

uniform mat4 p3d_ProjectionMatrix;
uniform mat4 p3d_ViewProjectionMatrix;
// uniform mat4 view_projection_matrix_inverse;

uniform sampler2D tex;
uniform sampler2D dtex;
// uniform sampler2D ntex;
in vec2 uv;
out vec4 color;

vec3 reconstructPosition(in vec2 uv, in float z)
{
    float x = uv.x * 2.0f - 1.0f;
    float y = (1.0 - uv.y) * 2.0f - 1.0f;
    vec4 position_s = vec4(x, y, z, 1.0f);
    mat4x4 view_projection_matrix_inverse = inverse(p3d_ViewProjectionMatrix);
    vec4 position_v = view_projection_matrix_inverse * position_s;
    return position_v.xyz / position_v.w;
}
// vec3 reconstructPosition(in vec2 uv, float depth) {
//     float z = depth * 2.0 - 1.0;
//
//     vec4 clipSpacePosition = vec4(uv * 2.0 - 1.0, z, 1.0);
//     vec4 viewSpacePosition = inverse(p3d_ViewProjectionMatrix) * clipSpacePosition;
//
//     // Perspective division
//     viewSpacePosition /= viewSpacePosition.w;
//
//     vec4 worldSpacePosition = inverse(p3d_ViewProjectionMatrix) * viewSpacePosition;
//
//     return worldSpacePosition.xyz;
// }

vec3 get_normal(vec2 texcoords) {
    const vec2 offset1 = vec2(0.0, 0.001);
    const vec2 offset2 = vec2(0.001, 0.0);

    float depth = texture(dtex, texcoords).r;
    float depth1 = texture(dtex, texcoords + offset1).r;
    float depth2 = texture(dtex, texcoords + offset2).r;

    vec3 p1 = vec3(offset1, depth1 - depth);
    vec3 p2 = vec3(offset2, depth2 - depth);

    vec3 normal = cross(p1, p2);
    normal.z = -normal.z;

    return normalize(normal);
}

// uniform float gSampleRad;
uniform mat4 gProj;
const int MAX_KERNEL_SIZE = 128;
uniform vec3 gKernel[MAX_KERNEL_SIZE];

void main() {
    float z = texture(dtex, uv).r;
    // depth = depth*depth;
    vec3 position = reconstructPosition(uv, z);
    vec3 n = get_normal(uv);
    // n = normalize(cross(dFdx(position), dFdy(position)));
    // n.z *= -1;
    // n.y *= -1;
    // n = n * vec3(1, -1, 1);
    // n = vec3(n.xyz * 0.5f + 0.5f) * .5;

    float AO = 0.0;
    float gSampleRad = .05;

    for (int i = 0 ; i < MAX_KERNEL_SIZE ; i++) {
        vec3 samplePos = position + gKernel[i]; // generate a random point
        vec4 offset = vec4(samplePos, 1.0); // make it a 4-vector
        offset = gProj * offset; // project on the near clipping plane
        offset.xy /= offset.w; // perform perspective divide
        offset.xy = offset.xy * 0.5 + vec2(0.5); // transform to (0,1) range

        // float sampleDepth = texture(gPositionMap, offset.xy).b;
        float sampleDepth = texture(dtex, offset.xy).b;

        if (abs(position.z - sampleDepth) < gSampleRad) {
            AO += step(sampleDepth,samplePos.z);
        }
    }

    AO = 1.0 - AO/128.0;

    color = vec4(pow(AO, 2.0));
}
''')



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    camera.shader = curvature_shader
    camera.clip_plane_near = 4
    camera.clip_plane_far = 500
    # camera.set_shader_input('contrast', 1)
    EditorCamera()
    #
    # def input(key):
    #     if key == 'space':
    #         if hasattr(camera, '_shader') and camera.shader:
    #             camera.shader = None
    #         else:
    #             camera.shader = curvature_shader
    # # from ursina.shaders import matcap_shader

    random.seed(2)
    for i in range(20):
        e = Entity(model='cube', position=Vec3(random.random(),random.random(),random.random())*2, rotation=Vec3(random.random(),random.random(),random.random())*360)
        # e.shader = matcap_shader
        # e.texture='blender_matcap'
        # e.model.generate_normals()


    app.run()



from ursina import *


class TransformGizmo(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'transform_gizmo'

        self.model = 'cube'
        self.scale *= .1
        self.color = color.lime

        self.add_to_selection = False
        self.entity_to_scale = None
        self.dist_to_cam = 0
        self.tool = 'none'
        self.tools = {
            'q' : 'none',
            'w' : 'move',
            'e' : 'rotate',
            'r' : 'scale'
            }

        self.button = None
        self.selection_buttons = list()

        self.prev_positions = list()
        self.original_parent = list()
        self.trash_list = list()


    # @undoable
    def move_entities(self, entities):
        # save these for undo
        # self.selection_copy = [e for e in entities]
        # self.prev_positions = [e.position for e in entities]
        self.temp_delta = mouse.delta
        self.prev_positions.append([e.position for e in entities])


        for e in entities:
            e.position = (
                e.position[0] + (mouse.delta[0] * self.dist_to_cam * camera.aspect_ratio),
                e.position[1] + (mouse.delta[1] * self.dist_to_cam),
                e.position[2])
            e.position = (round(e.x, 2), round(e.y, 2), round(e.z, 2))

        # print(entities[0].name, 'from:', self.prev_positions[-1][0], 'to:', entities[0].position)

        # undo
        yield 'move selected'
        print('redo move')
        for i, e in enumerate(entities):
            e.position = self.prev_positions[-1][i]

        self.prev_positions.pop()
        print(len(self.prev_positions))


    # @undoable
    def delete_selected(self):
        self.trash_list.append([e for e in scene.editor.selection])
        print(self.trash_list[-1])
        for e in self.trash_list[-1]:
            e.parent_before_destroyed = e.parent
            e.enabled = False
            e.parent = None
            print('m')

        scene.editor.hierarchy_panel.populate()
        print(self.trash_list)

        yield 'delete selected'
        print('restore detached')
        # get prev from trash can
        for e in self.trash_list[-1]:
            e.parent = e.parent_before_destroyed
            e.enabled = True
            # scene.editor.selection.append(e)

        del self.trash_list[-1]
        print('trash list is now:', self.trash_list)

        scene.editor.hierarchy_panel.populate()


    def update(self):
        if not scene.editor.enabled:    # just to make sure
            return
        # if mouse.hovered_entity.is_editor:
        #     return
        # for moving stuff in side view
        if (scene.editor.editor_camera.camera_pivot.rotation == (0,0,0)
        and mouse.hovered_entity
        and mouse.hovered_entity.is_editor == False
        and mouse.left):

            if mouse.delta_drag[0] != 0 or mouse.delta_drag[1] != 0:
                self.dist_to_cam = distance(
                    mouse.hovered_entity.get_pos(camera.render),
                    camera._cam.get_pos(camera.render)) * .2

                for i, e in enumerate(scene.editor.selection):
                    e.position = (
                        self.position[0] + (mouse.delta[0] * self.dist_to_cam * camera.aspect_ratio),
                        self.position[1] + (mouse.delta[1] * self.dist_to_cam),
                        self.position[2])
                    e.position = (round(e.x, 2), round(e.y, 2), round(e.z, 2))


# scale
        if self.entity_to_scale:
            self.entity_to_scale.scale = self.entity_original_scale * (1 + (mouse.x - self.start_mouse) * 4)
            self.entity_to_scale.scale = [round(s, 2) for s in self.entity_to_scale.scale]



    def input(self, key):
        if key == 'left mouse down':
            if mouse.hovered_entity and mouse.hovered_entity.is_editor:
                return
            if not mouse.hovered_entity:
                scene.editor.selection.clear()

            elif mouse.hovered_entity and mouse.hovered_entity.is_editor == False:
                # select entities
                if not self.add_to_selection:
                    scene.editor.selection.clear()
                    scene.editor.selection.append(mouse.hovered_entity)
                else:
                    scene.editor.selection.clear()

                if len(scene.editor.selection) > 0:
                    self.position = scene.editor.selection[-1].world_position
                    self.start_positions = [e.position for e in scene.editor.selection]
                    self.original_parents = [e.parent for e in scene.editor.selection]
                    for e in scene.editor.selection:
                        e.reparent_to(scene)


        if key == 'left mouse up':
            if mouse.hovered_entity and mouse.hovered_entity.is_editor:
                return

            for i, e in enumerate(scene.editor.selection):
                # self.original_parent = e.position
                # e.reparent_to(scene)
                e.parent = self.original_parents[i]
                e.position = self.start_positions[i]
                # e.reparent_to(self.original_parent)

            self.move_entities(scene.editor.selection)
            if len(scene.editor.selection) > 0:
                self.position = scene.editor.selection[-1].world_position
            # scene.editor.selection.clear()



        if key == 'left shift':
            self.add_to_selection = True
        if key == 'left shift up':
            self.add_to_selection = False


        if key == 's' and not held_keys['control']:
            if scene.editor.selection[0]:
                self.start_mouse = mouse.x
                self.entity_to_scale = scene.editor.selection[0]
                print(self.entity_to_scale)
                self.entity_original_scale = self.entity_to_scale.scale
        if key == 's up':
            # self.entity_to_sale
            self.entity_to_scale = None

# selection buttons
        if key == 't':
            entities = list(scene.entities)
            for e in entities:
                if not e.is_editor and e is not camera and e is not scene.ui:
                    self.button = EditorButton()
                    self.button.is_editor = True
                    self.button.parent = render
                    self.button.position = e.world_position
                    # self.button.position *= 2
                    # self.button.scale *= .01
                    self.button.text = e.name
                    self.button.look_at(camera)
                    # self.button.rotation_y += 180
                    # self.button.rotation_z += 180
                    # self.button.scale_z -= 1
                    # self.button.rotation = camera.rotation
                    # self.button.text_entity.scale *= .5
                    self.button.color = color.orange
                    self.button_script = self.button.add_script('selection_button')
                    self.button_script.selection_target = e
                    self.selection_buttons.append(self.button)

        if key == 't up':
            for b in self.selection_buttons:
                destroy(b)
            self.button = None
            self.button_script = None
            self.selection_buttons.clear()


        self.tool = self.tools.get(key, self.tool)

        if key == 'delete':
            self.delete_selected()


if __name__ == '__main__':
    app = Ursina()
    TransformGizmo()
    app.run()



from panda3d.core import Shader


camera_normals_shader = Shader.make('''

void vshader(float4 vtx_position : POSITION,
             out float4 l_position : POSITION,
             out float2 l_texcoord : TEXCOORD0,
             out float2 l_texcoordD : TEXCOORD1,
             out float2 l_texcoordN : TEXCOORD2,
             uniform float4 texpad_depth,
             uniform float4 texpad_normal,
             uniform float4x4 mat_modelproj)
{
  l_position = mul(mat_modelproj, vtx_position);
  l_texcoord = vtx_position.xz;
  l_texcoordD = (vtx_position.xz * texpad_depth.xy) + texpad_depth.xy;
  l_texcoordN = (vtx_position.xz * texpad_normal.xy) + texpad_normal.xy;
}


void fshader(out float4 o_color : COLOR,
             uniform float4 k_params1,
             uniform float4 k_params2,
             float2 l_texcoord : TEXCOORD0,
             float2 l_texcoordD : TEXCOORD1,
             float2 l_texcoordN : TEXCOORD2,
             uniform sampler2D k_random : TEXUNIT0,
             uniform sampler2D k_depth : TEXUNIT1,
             uniform sampler2D k_normal : TEXUNIT2)
{
    float4 c = tex2D(k_random, l_texcoord);
    o_color = c;
}

''', Shader.SL_Cg)



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere')
    e = Entity(model='cube', y=-1)
    camera.shader = camera_normals_shader
    EditorCamera()

    app.run()



from ursina import Shader; bloom_shader = Shader(
fragment='''
#version 430

uniform sampler2D tex;
in vec2 uv;
uniform float blur_size;
uniform float strength;
out vec4 color;
in vec2 window_size;

const int samples = 8;
const vec2 offsets[samples] = vec2[](vec2(1.0,0.0), vec2(.7,.7), vec2(1.,0.), vec2(.7,-.7), vec2(0.,-1), vec2(-.7,-1), vec2(-1.,0.), vec2(-.7,.7));

void main() {
    vec4 blurred_image = vec4(.0);
    float str = .01;
    // float aspect_ratio = window_size[0] / window_size[1];
    float aspect_ratio = 16./9.;

    for (int i=0; i<samples; i++) {
        // vec2 offset_uv = uv + vec2(0., (i/(samples-1) - 0.5) * blur_size*str);
        vec2 offset_uv = (uv + (offsets[i] * str * blur_size / vec2(aspect_ratio, 1.)));
        blurred_image += texture(tex, offset_uv) /samples/2;

        //offset_uv = uv + vec2((i/(samples-1) - 0.5) * blur_size*str / aspect_ratio, 0.);
        //blurred_image += texture(tex, offset_uv) /samples/2;
    }

    color = texture(tex, uv);
    vec4 a = mix(vec4(0.), blurred_image, strength);
    color += a;
    // color /= 2.;
    // color += blurred_image;
}
''',
default_input=dict(
    blur_size = .5,
    strength = 1,
    window_size = (1920/1080)

))

if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    window.color = color._16

    e = Entity(model='sphere', color=color.orange)
    e = Entity(model='cube', y=-1)
    e = Entity(model='plane', texture='grass', scale=10, y=-1)

    e = Entity(model='sphere', color=color.black, x=3)
    e = Entity(model='sphere', color=color._8, x=3, z=1, scale=2)
    camera.shader = bloom_shader


    slider = ThinSlider(max=1, dynamic=True, position=(-.25, -.45))

    def set_blur():
        print(slider.value)
        camera.set_shader_input("blur_size", slider.value)

    # def update():
    #     camera.set_shader_input('blur_size', mouse.x)


    slider.on_value_changed = set_blur
    EditorCamera()

    app.run()



from ursina import *
from pathlib import Path
# app = Ursina()


# img = load_texture('background_header.png')
# Entity(model='quad', texture=img, scale=[img.width/100, img.height/100])
# bg = Entity(model='quad', scale=[img.width/100, img.height/100], z=.1)
#
# for y in range(img.height):
#     for x in range(img.width):
#         col = img.get_pixel(x,y)
        # print(col)
        # new_color = hsv(col.h, col.s, col.v)
        # img.set_pixel()

# img.apply()
# app.run()
from ursina.ursinamath import chunk_list
from PIL import Image
#
# path = Path('.').parent.parent.parent / 'docs' / 'background_header.png'
# print(path)
img = Image.open('background_header.png')
palette = img.getpalette()
# print(palette)
palette = list(chunk_list(palette, 3))
w, h = img.size

for y in range(h):
    for x in range(w):
        palette_index = img.getpixel((x, y))
        col = palette[palette_index]
        col = color.rgb(*col)
        print(col)



from ursina import Entity, Button, camera, color, Text, window, mouse, destroy, Vec2
from ursina.scripts.grid_layout import grid_layout
from ursina.scripts.property_generator import generate_properties_for_class


@generate_properties_for_class()
class ButtonGroup(Entity):
    default_selected_color = color.azure
    default_highlight_selected_color = default_selected_color

    def __init__(self, options, default='', min_selection=1, max_selection=1, origin=(-.5,.5), spacing=(0.025,0), label='', max_x=99, **kwargs):
        super().__init__()
        self.buttons = []
        self.selected = []

        self.deselected_color =         Button.default_color
        self.highlight_color =          Button.default_highlight_color if Button.default_highlight_color is not None else Button.default_color.tint(.2)
        self.selected_color =           ButtonGroup.default_selected_color
        self.highlight_selected_color = ButtonGroup.default_highlight_selected_color

        self.min_selection = min_selection
        self.max_selection = max(min_selection, max_selection)
        self.origin = origin
        self.spacing = spacing
        self.max_x = max_x
        self.label_text_entity = None
        if label:
            self.label = label

        self.options = options

        self.parent = camera.ui
        self.scale = Text.size * 2


        for key, value in kwargs.items():
            setattr(self, key, value)

        if default:
            for b in [e for e in self.buttons if e.value == default]:
                self.select(b)
        else:
            for i in range(min_selection):
                self.select(self.buttons[i])


    def options_setter(self, value):
        self._options = value
        self.layout()


    def origin_setter(self, value):
        if not isinstance(value, Vec2):
            value = self._list_to_vec(value)

        self._origin = value
        if not self.buttons:
            return
        self.layout()

    def value_getter(self):
        if self.max_selection == 1:
            return self.selected[0].value

        return [b.value for b in self.selected]

    def value_setter(self, value):
        [self.select(b) for b in self.buttons if b.value in value]


    def label_setter(self, value):
        if not self.label_text_entity:
            self.label_text_entity = Text(parent=self, world_scale=25/2, origin=(0,-.5), position=(0,0,-1), color=color.text_color)
        self.label_text_entity.text = value


    def layout(self):
        [destroy(c) for c in self.buttons]
        self.buttons = []
        longest_word = max(self.options, key=len) + '__' # padding
        width = Text.get_width(longest_word) / Text.size / 2

        for e in self.options:
            b = Button(parent=self, text=e, name=e, scale_x=width, scale_y=.9, highlight_scale=1, pressed_scale=1)
            b.value = e
            self.buttons.append(b)

        grid_layout(self.buttons, origin=self.origin.xy, spacing=self.spacing, max_x=self.max_x)
        if self.label_text_entity:
            self.label_text_entity.origin = Vec2(self.origin.x, -self.origin.y)
            self.label_text_entity.position = Vec2.zero
            if self.origin_y == 0:
                self.label_text_entity.origin = Vec2(-self.origin.x, 0)
            if self.origin == Vec2.zero:
                self.label_text_entity.x = -(width * min(len(self.buttons), self.max_x) * .5)
                self.label_text_entity.origin = Vec2(.5,0)


    def input(self, key):
        if key == 'left mouse down' and mouse.hovered_entity in self.buttons:
            self.select(mouse.hovered_entity)


    def select(self, b):
        if b in self.selected and self.min_selection > 0 and len(self.selected) >= self.min_selection:
            return

        # add
        if b not in self.selected:
            b.color = self.selected_color
            b.highlight_color = self.highlight_selected_color
            self.selected.append(b)

            if len(self.selected) > self.max_selection:
                # remove oldest addition
                self.selected[0].color = self.deselected_color
                self.selected[0].highlight_color = self.highlight_color
                self.selected.pop(0)

        self.on_value_changed()



    def on_value_changed(self): # assign a function to this to make something happen when you change the ButtonGroup's value
        pass



if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    # # test setting custom default colors
    # Text.default_font = 'VeraMono.ttf'
    # color.text_color = color.orange
    # Button.default_highlight_color = color.blue
    # Button.default_color = color.turquoise
    # ButtonGroup.default_highlight_selected_color = color.white

    center = Entity(parent=camera.ui, model='circle', scale=.005, color=color.red, z=-1)
    gender_selection = ButtonGroup(('man', 'woman', 'other'),
        origin=(-.5,0),
        label='choose gender:',
        max_x=1
        )

    def on_value_changed():
        print('set gender:', gender_selection.value)
    gender_selection.on_value_changed = on_value_changed


    window.color = color._32

    # test
    for e in [(-.5,.5), (0,.5), (.5,.5), (-.5,0), (0,0), (.5,0), (-.5,-.5), (0,-.5), (.5,-.5)]:
        Button(
            text='*',
            model='quad',
            text_origin=e,
            scale=.095,
            origin=(-.5,.5),
            position = window.top_left + Vec2(*e)*.2 + Vec2(.1,-.1),
            tooltip=Tooltip(str(e)),
            on_click=Func(setattr, gender_selection, 'origin', e),
        )

    app.run()



from ursina.entity import Entity
from ursina.ursinamath import distance
import time

# TODO:
# circle
# AArectangle
# AAbox
# box

# sphere

class Trigger(Entity):
    def __init__(self, **kwargs):
        super().__init__()

        self.trigger_targets = None
        self.radius = .5
        self.triggerers = []
        self.update_rate = 4
        self._i = 0

        for key, value in kwargs.items():
            setattr(self, key, value)


    def update(self):
        self._i += 1
        if self._i < self.update_rate:
            return

        self._i = 0

        for other in self.trigger_targets:
            if other == self:
                continue

            dist = distance(other, self)
            if not other in self.triggerers and dist <= self.radius:
                self.triggerers.append(other)
                if hasattr(self, 'on_trigger_enter'):
                    self.on_trigger_enter()
                continue

            if other in self.triggerers and dist > self.radius:
                self.triggerers.remove(other)
                if hasattr(self, 'on_trigger_exit'):
                    self.on_trigger_exit()
                continue

            if other in self.triggerers and hasattr(self, 'on_trigger_stay'):
                self.on_trigger_stay()




if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    player = Entity(model='cube', color=color.azure, scale=.05)
    def update():
        player.x += held_keys['d'] * time.dt * 2
        player.x -= held_keys['a'] * time.dt * 2

    t = Trigger(trigger_targets=(player,), x=1, model='sphere', color=color.hsv(0,1,1,.5))
    t.on_trigger_enter = Func(print, 'enter')
    t.on_trigger_exit =  Func(print, 'exit')
    t.on_trigger_stay =  Func(print, 'stay')

    app.run()



from ursina import *
from ursina.prefabs.first_person_controller import FirstPersonController


app = Ursina()

ground = Entity(model='plane', scale=(100,1,100), color=color.yellow.tint(-.2), texture='white_cube', texture_scale=(100,100), collider='box')

e = Entity(model='cube', scale=(1,5,10), x=2, y=.01, rotation_y=45, collider='box', texture='white_cube')
e.texture_scale = (e.scale_z, e.scale_y)

e = Entity(model='cube', scale=(1,5,10), x=-2, y=.01, collider='box', texture='white_cube')
e.texture_scale = (e.scale_z, e.scale_y)

player = FirstPersonController(model='cube', y=1, origin_y=-.5)
player.gun = None

gun = Button(parent=scene, model='cube', color=color.blue, origin_y=-.5, position=(3,0,3), collider='box')
gun.on_click = Sequence(Func(setattr, gun, 'parent', camera), Func(setattr, player, 'gun', gun))

gun_2 = duplicate(gun, z=7, x=8)
slope = Entity(model='cube', collider='box', position=(0,0,8), scale=6, rotation=(45,0,0), texture='brick', texture_scale=(8,8))
slope = Entity(model='cube', collider='box', position=(5,0,10), scale=6, rotation=(80,0,0), texture='brick', texture_scale=(8,8))
Text('press tab to toggle slow motion', y=.5)

def input(key):
    if key == 'left mouse down' and player.gun:
        gun.blink(color.orange)
        bullet = Entity(parent=gun, model='cube', scale=.1, color=color.black)
        bullet.world_parent = scene
        bullet.animate_position(bullet.position+(bullet.forward*50), curve=curve.linear, duration=1)
        destroy(bullet, delay=1)

    if key == 'tab':
        # application.paused = not application.paused
        if application.time_scale == 1:
            application.time_scale = .3
        else:
            application.time_scale = 1



app.run()



from ursina import *



app = Ursina()
Texture.default_filtering = 'trilinear'
# application.asset_folder = Path('C:/sync/3d')
from ursina.shaders import lit_with_shadows_shader

# from ursina.shaders import matcap_shader
# Entity(model='rock_scene', shader=matcap_shader, texture='shore')



from ursina.shaders import triplanar_shader
triplanar_shader.default_input['side_texture'] = load_texture('grass', application.internal_textures_folder)
triplanar_shader.default_input['side_texture_scale'] = Vec2(1/10,1/10)

e = Entity(model='floating_isle', shader=triplanar_shader, texture='brick', texture_scale=Vec2(.1))
EditorCamera()

# PointLight(y=3)
sun = DirectionalLight(y=50, rotation_x=120-90)
# sun._light.get_lens().set_near_far(0,30)
#
# sun._light.show_frustum()
def input(key):
    if key == 'escape':
        quit()


app.run()



from ursina import *
from ursina.scripts.property_generator import generate_properties_for_class
from ursina.shaders.unlit_shader import unlit_shader

# _instances = dict() # keeps track of all the instances so we can make a menu that lists all instances, with the option for baking them.

cache = dict()
# If a name and seed is provided, it will replay the instance if it exists.
# This will be faster to play since you only have to instantiate the paritcles and animations once.
# However, this means you can only have one instance playing at the same time.
# If the particle effect has been baked, it will instead load an FrameAnimation3D.



def play_particle_system(name, use_cache=True, auto_play=True, auto_destroy=True, unscaled=False, **kwargs):
    # print('try loading particle system:', name)
    if name in cache:
        animation_texture = cache[name]
    else:
        animation_texture = load_texture(f'{name}_*_baked_fps*_bounds*.png')
        cache[name] = animation_texture

    if animation_texture:   # if the particle system has been baked
        vat_instance = vertex_animation(animation_texture, auto_play=auto_play, unscaled=unscaled)
        if not auto_play:
            vat_instance.enabled = False

        for key, value in kwargs.items():
            setattr(vat_instance, key, value)
        return vat_instance

    whole_file_name = f'{name}.ursinaparticles'
    particle_files = tuple(application.asset_folder.glob(f'**/{whole_file_name}'))
    if not particle_files:
        print_warning('missing particle system:', whole_file_name)
        return

    with particle_files[0].open('r') as f:
        text = f.read()

    relevant_part = 'ParticleSystemContainer(' + text.split('ParticleSystemContainer(',1)[1].split("if __name__ == ")[0]
    particle_container = eval(relevant_part)
    for key, value in kwargs.items():
        setattr(particle_container, key, value)

    for subsystem in particle_container.subsystems:
        for seq in subsystem.anims:
            seq.unscaled = unscaled

    if auto_play:
        particle_container.play()
    if auto_destroy:
        destroy(particle_container, delay=particle_container.total_duration, unscaled=unscaled)

    return particle_container


class ParticleSystemContainer(Entity):
    instances = []

    def __init__(self, subsystems:list, **kwargs):
        super().__init__(**kwargs)
        self.subsystems = subsystems
        for e in self.subsystems:
            e.parent = self
        self.total_duration = max([c.delay + c.total_duration for c in self.subsystems])


    def play(self):
        for e in self.subsystems:
            e.play()


    def bake_to_vertex_animation_texture(entity, name, seed, fps=30):   # bakes all children and animations to a texture
        animations = []
        for c in entity.get_decendants():
            animations.extend(c.anims)

        frames = []
        bake_parent = Entity()
        original_parent = entity.parent
        entity.parent = bake_parent

        duration_per_frame = 1 / fps
        total_duration = max(seq.func_call_time[-1] for seq in entity.animations)
        num_frames = int(total_duration / duration_per_frame)
        if num_frames <= 0:
            raise Exception('can\'t bake particle system with 0 frames')
        print('num_frames:', num_frames)


        for i in range(num_frames):
            bake_parent.model = None
            if not bake_parent.children:
                raise Exception('no particles to bake')

            bake_parent.combine(auto_destroy=False)

            for seq in animations:
                seq.t += duration_per_frame
                seq.started = True
                seq.update()
                seq.started = False

            frames.append(deepcopy(bake_parent.model))
            Entity(model=Mesh(vertices=bake_parent.model.vertices), wireframe=True, color=color.green) # debug render

        entity.parent = original_parent



    def bake(self, name, seed, fps=30):     # bake to vertex animation texture
        from PIL import Image

        animations = []
        for c in self.subsystems:
            animations.extend(c.anims)
            # print('------------',c.total_duration)

        duration_per_frame = 1 / fps
        num_frames = int(self.total_duration / duration_per_frame)
        if num_frames <= 0:
            raise Exception('can\'t bake particle system with 0 frames')

        print('num_frames:', num_frames)

        folder = application.asset_folder / 'particle_system_baked'
        folder.mkdir(parents=True, exist_ok=True)
        frames = []
        bake_parent = Entity()
        original_parent = self.parent
        self.parent = bake_parent

        for i in range(num_frames):
            bake_parent.model = None
            if not self.children:
                raise Exception('no particles to bake')

            bake_parent.combine(auto_destroy=False)

            for seq in animations:
                seq.t += duration_per_frame
                seq.started = True
                seq.update()
                seq.started = False

            frames.append(deepcopy(bake_parent.model))
            # print(bake_parent.model.vertices)
            Entity(model=Mesh(vertices=bake_parent.model.vertices), wireframe=True, color=color.green) # debug render

        self.parent = original_parent

        if not frames:
            raise Exception('Particle system has 0 frames')

        # num_vertices = max(len(e.vertices) for e in frames)
        num_vertices = len(frames[0].vertices)
        if num_vertices <= 0:
            raise Exception('can\'t bake particle system with 0 vertices.')

        min_x, max_x, min_y, max_y, min_z, max_z = 0, 0, 0, 0, 0, 0
        for frame in frames:
            min_x = min(min_x, floor(min(v.x for v in frame.vertices)))
            min_y = min(min_y, floor(min(v.y for v in frame.vertices)))
            min_z = min(min_z, floor(min(v.z for v in frame.vertices)))
            max_x = max(max_x, ceil(max(v.x for v in frame.vertices)))
            max_y = max(max_y, ceil(max(v.y for v in frame.vertices)))
            max_z = max(max_z, ceil(max(v.z for v in frame.vertices)))

        # print('verts:', num_vertices)
        texture_width = num_vertices * 2    # one pixel per vertex. one color per vertex.
        texture_height = len(frames)

        if texture_width > 4096 or texture_height > 4096:
            raise Exception(f'Particle system animation won\'t fit in a 4096x4096 texture: num_frames:{len(frames.vertices)}, vertices:{num_vertices}')

        print(texture_width, texture_height)
        texture = Texture(Image.new(mode="RGB", size=(texture_width, texture_height)))

        for y, frame in enumerate(frames):
            for x, v in enumerate(frame.vertices):
                # v = v - frames[0][x] # get offset from first frame
                texture.set_pixel(x, y, color.rgb(
                    inverselerp(min_x, max_x, v.x),
                    inverselerp(min_y, max_y, v.y),
                    inverselerp(min_z, max_z, v.z))
                )
            for x, c in enumerate(frame.colors):
                texture.set_pixel(x+num_vertices, y, c)

        # Mesh(vertices=frames[0]).save(f'{self.name}_{self.seed}_baked_{min_x}_{max_x}_{min_y}_{max_y}_{min_z}_{max_z}.ursinamesh', folder)
        destroy(bake_parent)

        texture.save(folder / f'{name}_seed{seed}_baked_fps{fps}_bounds{min_x}_{max_x}_{min_y}_{max_y}_{min_z}_{max_z}.png')


def vertex_animation(animation_texture, auto_play=True, unscaled=False):
    from ursina.shaders.vertex_animation_shader import vertex_animation_shader
    instance = Entity(model=Mesh(vertices=[Vec3.zero for i in range(animation_texture.width)]), shader=vertex_animation_shader)

    fps = float(animation_texture.path.stem.split('_fps')[1].split('_')[0])
    min_x, max_x, min_y, max_y, min_z, max_z = [int(e) for e in animation_texture.path.stem.split('_bounds')[1].split('_')]
    animation_texture.filtering = 'bilinear'
    animation_texture.repeat = False

    instance.set_shader_input('pos_min', Vec3(min_x, min_y, min_z))
    instance.set_shader_input('pos_max', Vec3(max_x, max_y, max_z))
    instance.set_shader_input('num_verts', animation_texture.width)
    instance.set_shader_input('frame_texture', animation_texture)
    instance.set_shader_input('total_frames', animation_texture.size[1])
    animation_duration = 1/fps * animation_texture.size[1]
    loop = False

    if auto_play:
        instance.animate_shader_input('frame_index', animation_texture.size[1]-1, duration=animation_duration, loop=loop, curve=curve.linear, resolution=120, unscaled=unscaled)

    if not loop:
        destroy(instance, delay=animation_duration, unscaled=unscaled)
    return instance

def _sample_random(seq, i):
    return random.choice(seq)

def _sample_sequential(seq, i):
    if i >= len(seq):
        raise IndexError(f'{i} out of range. Make sure start_color and end_color lists are the same length as spawn_points when using color_sample_function=\'sequential\'')
    return seq[i]

color_sample_functions = {
    'random' : _sample_random,
    'sequential' : _sample_sequential,
}


@generate_properties_for_class()
class ParticleSystem(Entity):
    default_values = dict(
        start_size=1,
        end_size=0,
        size_curve=curve.in_expo,
        color_curve=curve.in_expo,

        start_color=color.white,
        end_color=color.white,
        color_sample_function='random',    #  'random' / 'sequential'

        start_direction=Vec3(0,0,0),
        direction_randomness=Vec3(0,0,0),
        start_rotation=Vec3(0,0,0),
        rotation_randomness=Vec3(0,0,0),    # rotates the model only, not the direction it moves in
        spin=Vec3(0,0,0),
        spin_curve = curve.linear,

        num_particles=1,   # if 0, will be set to the length of the spawn_points list
        speed=0,
        speed_curve=curve.linear,
        move_directions='forward',  # local forward

        bounce=0,
        bounce_curve=curve.out_bounce,

        max_particles=20,
        lifetime=1,
        mesh='quad',
        double_sided=True,

        world_space=False,
        shader=unlit_shader,
        texture=None,
        always_on_top=False,
        unlit=True,

        spawn_points=(Vec3.zero, ),
        spawn_type='burst', # 'burst', 'random' / 'sequential' / 'continious'
        loop_every=0,
        spawn_interval=0,     # only applies if spawn_type is random, sequential or continuous
        delay=0,
        auto_play=False,
        # auto_destroy=True,
        name='',
        seed=None,  # set which random seed to use. int / int tuple/list
    )

    def __init__(self, **kwargs):
        self.original_settings = kwargs

        # invalid_keys = set(kwargs) - set(Entity.default_values | __class__.default_values)
        # if invalid_keys:
        #     raise ValueError(f"Invalid keyword arguments: {', '.join(invalid_keys)}")

        super().__init__(**(__class__.default_values | kwargs))

        if self.num_particles == 0:
            self.num_particles = len(self.spawn_points)

        if not isinstance(self.move_directions, tuple | list):
            self.move_directions = [self.move_directions for i in range(self.num_particles)]
        if not isinstance(self.start_direction, tuple | list):
            self.start_direction = [self.start_direction for i in range(self.num_particles)]

        if not isinstance(self.start_color, tuple | list):
            self.start_color = (self.start_color, )
        if not isinstance(self.end_color, tuple | list):
            self.end_color = (self.end_color, )

        if not isinstance(self.speed, tuple | list):
            self.speed = (self.speed, self.speed)

        self.spawn_points = LoopingList(self.spawn_points)
        self.particles = []
        self.t = 0
        self.total_duration = self.lifetime + (self.num_particles * self.spawn_interval)
        self.is_playing = False

        self.generate()
        if self.auto_play:
            self.play()


    def update(self):
        if self.loop_every == 0:
            return
        if self.is_playing:
            # print(self.t,)
            self.t += time.dt

        if self.t > self.loop_every:
            self.generate()
            self.play()
            self.t = 0


    def generate(self):
        self.anims = []

        if self.spawn_type == 'burst':
            [self.generate_particle_animations(
                position=Vec3.zero,
                start_direction=self.start_direction[i],
                move_direction=self.move_directions[i],
                delay=self.delay+(i*self.spawn_interval), i=i)
                for i in range(self.num_particles)]

        elif self.spawn_type == 'random':
            [self.generate_particle_animations(
                position=random.choice(self.spawn_points),
                start_direction=self.start_direction[i],
                move_direction=self.move_directions[i],
                delay=self.delay+(i*self.spawn_interval), i=i)
                for i in range(self.num_particles)]

        elif self.spawn_type == 'sequential':
            [self.generate_particle_animations(
                position=self.spawn_points[i],
                start_direction=self.start_direction[i],
                move_direction=self.move_directions[i],
                delay=self.delay+(i*self.spawn_interval), i=i)
                for i in range(self.num_particles)]


    def play(self):
        self.t = 0
        for e in self.anims:
            e.start()
        self.is_playing = True


    def generate_particle_animations(self, position, start_direction, move_direction, delay=0, i=0):
        if self.seed is None:
            random.seed(None)
        else:
            random.seed(self.seed+i)    # use seed + particle index so the particle system will be somewhat random, but prevent each particle from facing the same direction.

        if len(self.particles) >= self.max_particles:
            return

        model = self.mesh
        if not isinstance(model, str):
            model = deepcopy(model)

        e = Entity(parent=self, scale=self.start_size, position=position, enabled=False)
        # e.rotation = start_direction + Vec3(*[random.uniform(-e/2, e/2) for e in self.direction_randomness])
        e.look_in_direction(start_direction)
        e.rotation += Vec3(*[random.uniform(-e/2, e/2) for e in self.direction_randomness])

        if isinstance(move_direction, str):
            move_direction = getattr(e, move_direction).normalized()
            # e.look_at(e.position + direction)
        graphics = Entity(parent=e, model=model, origin=self.origin, double_sided=self.double_sided, shader=self.shader, texture=self.texture, always_on_top=self.always_on_top, unlit=self.unlit)
        e.graphics = graphics
        graphics.rotation = self.start_rotation + Vec3(*[random.uniform(-e/2, e/2) for e in self.rotation_randomness])

        if self.world_space:
            e.world_parent = scene

        # if self.color_randomness ==
        graphics.color = color_sample_functions[self.color_sample_function](self.start_color, i)
        should_destroy_particles = self.loop_every > 0

        self.anims.append(Sequence(
            Func(self.try_disabling, e),
            Wait(delay),
            Func(self.try_enabling, e),
            Func(setattr, e, 'position', e.position),
            Func(setattr, e, 'rotation', e.rotation),
            Func(setattr, e, 'scale', e.scale),
            Func(setattr, graphics, 'color', graphics.color),
            Func(setattr, graphics, 'rotation', graphics.rotation),
            # Func(setattr, graphics, 'scale', 1),
            Func(setattr, graphics, 'y', graphics.y),
            started=False, auto_destroy=should_destroy_particles, name='start_sequence'))

        RESOLUTION = int(self.lifetime * 60)
        self.anims.extend(e.animate_position(e.position + (move_direction * random.uniform(self.speed[0], self.speed[1])), duration=self.lifetime, delay=delay, curve=self.speed_curve, resolution=RESOLUTION, auto_play=False, auto_destroy=should_destroy_particles))
        if self.end_size != self.start_size:
            self.anims.append(e.animate_scale(self.end_size, duration=self.lifetime, delay=delay, curve=self.size_curve, resolution=RESOLUTION, auto_play=False, auto_destroy=should_destroy_particles))

        if self.end_color != self.start_color:
            self.anims.append(graphics.animate_color(color_sample_functions[self.color_sample_function](self.end_color,i), duration=self.lifetime, delay=delay, curve=self.color_curve, resolution=RESOLUTION, auto_play=False, auto_destroy=should_destroy_particles))

        if self.spin:
            self.anims.extend(graphics.animate_rotation(self.spin * self.lifetime, duration=self.lifetime, delay=delay, curve=self.spin_curve, resolution=RESOLUTION, auto_play=False, auto_destroy=should_destroy_particles))

        self.anims.append(graphics.animate_y(self.bounce, duration=self.lifetime, delay=delay, curve=self.bounce_curve, resolution=RESOLUTION, auto_play=False, auto_destroy=should_destroy_particles))
        # graphics.animate_y(self.sway, duration=self.lifetime, delay=delay, curve=self.sway_curve)

        self.anims.append(
            Sequence(
                Wait(self.lifetime+delay),
                Func(self.try_disabling, e),
                started=False,
                auto_destroy=should_destroy_particles,
                name='end_sequence',
            )
        )
        if self.loop_every > 0:
            self.anims.append(Sequence(
                Wait(delay + self.total_duration + .1),
                Func(destroy, e),
                auto_destroy=True, name='destroy_particle_sequence')
                )

        # if self.loop_every > 0:
        #     destroy(self, delay=delay + self.total_duration + .1)
            # del cache[self.name]
        #     self.anims.append(Sequence(Func(destroy, self, delay=self.lifetime+delay+.05)))
        # if not self.loop_every and not cached:
        # random.seed(None)
        # self.anims.extend(animations)
        # print('-----', delay, '-------', animations)
        # return animations
        # print('num anims:', len(self.anims))

    # def destroy(self):
    #     if self.name and self.seed is not None:
    #         id = f'{self.name}_{self.seed}'
    #         if id in cache:
    #             del cache[id]

    #     destroy(self)

    def try_enabling(self, particle):
        try:
            particle.enabled = True
            particle.model_parent.scale = 1
        except:
            pass
            # print('can not enable particle because the entity has been destroyed.', self)


    def try_disabling(self, particle):
        try:
            # particle.enabled = False
            particle.model_parent.world_scale = Vec3(.001)  # keep enabled and scale to 0 to keep vertices after combine() and make baking easier.
        except:
            pass
            # print('can not disable particle, already destroyed')

# from ursina.prefabs.vec_field import VecField
# from ursina.editor.level_editor import ColorField
# class ParticleSystemUI(Entity):
#     def __init__(self):
#         super().__init__()
#         self.parent = camera.ui
#         self.ui = Entity(parent=self, enabled=1, position=window.top_left+Vec3(.1,-.1,0), scale=.75, model=Circle(radius=.01))
#         i = 0
#         for key ,value in _ParticleSystem.default_values.items():
#             # print(i, key ,value)
#             field = None
#             if isinstance(value, Vec3):
#                 field = VecField(value, parent=self.ui)

#             elif isinstance(value, bool):
#                 # print('---aaa')
#                 field = Checkbox(parent=self.ui, x=.025, value=value)


#             elif isinstance(value, int):
#                 field = VecField(value, parent=self.ui)

#             elif isinstance(value, Color):
#                 field = ColorField(attr_name='color', is_shader_input=False, value=value, parent=self.ui, scale_y=.05)

#             if field:
#                 field.y = -i * .05
#                 Text(text=key, parent=field, y=.5, z=-1, x=.5, world_scale=20)
#                 i += 1
#             else:
#                 print('no field for:', key, value)
#         print('made fields')

#     def input(self, key):
#         combo = ['control', 'shift' , 'alt', 'p', 'a']
#         if key in combo and all([held_keys[e] for e in combo]):
#             self.ui.enabled = not self.ui.enabled


class ParticleSystemUI(Entity):
    def __init__(self, asset_file, particle_system_container):
        super().__init__(parent=camera.ui)
        from ursina import Text
        from ursina.prefabs.window_panel import WindowPanel

        self.asset_file = asset_file
        self.particle_system_container = particle_system_container
        self.seed_slider = Slider(min=0, max=16, step=1, x=-.5-.1, y=-.4, text='seed')
        def preview_particle_system():
            particle_system_container.seed = self.seed_slider.value
            for c in self.particle_system_container.subsystems:
                c.seed = self.seed_slider.value
                c.generate()
            self.particle_system_container.play()

        self.seed_slider.on_value_changed = preview_particle_system

        self.play_button = Button(text='Play', scale=[.1,.05], y=-.4, on_click=preview_particle_system)
        self.ground = Entity(model='wireframe_quad', scale=8, rotation_x=90, alpha=.2)

        self.overwrite_label = Text('Overwrite\nFILE?', scale=.75)
        self.overwrite_button = Button(text='Yes', color=color.azure)
        self.cancel_button = Button(text='Cancel', color=color.black)

        self.overwrite_dialog = WindowPanel(
            title='Overwrite?',
            content=[self.overwrite_label, self.overwrite_button, self.cancel_button],
            parent=self, z=-10, popup=True, enabled=False)

        self.cancel_button.on_click = self.overwrite_dialog.disable
        self.play_baked_buttons = []
        self.bake_buttons = []
        self.render_particle_system_list()

    def input(self, key):
        if key == 'tab':
            self.render_particle_system_list()


    def on_enable(self):
        self.render_particle_system_list()


    def render_particle_system_list(self):
        [destroy(e) for e in self.play_baked_buttons]
        [destroy(e) for e in self.bake_buttons]
        self.play_baked_buttons = []
        self.bake_buttons = []

        vertex_animation_paths = tuple(application.asset_folder.glob(f'**/{self.asset_file.stem}_seed*_baked_fps*_bounds*.png'))

        for vat_path in vertex_animation_paths:
            fps = vat_path.name.split('_fps')[1].split('_',1)[0]
            play_baked_button = Button(parent=self, scale=(.2,.05), text=f'{vat_path.stem}\n(VAT) {fps}FPS', text_size=.5, color=color.green.tint(-.5))
            def play_baked(vat_path=vat_path):
                vertex_animation(Texture(vat_path))
            play_baked_button.on_click = play_baked
            self.play_baked_buttons.append(play_baked_button)

        for target_fps in (12,30,60):
            def _try_bake(target_fps=target_fps):
                animation_texture = load_texture(f'{self.asset_file.stem}_seed*_baked_fps*_bounds*.png')
                if not animation_texture:
                    self.particle_system_container.bake(self.asset_file.stem, self.seed_slider.value, target_fps)
                    return
                if animation_texture.path.exists():
                    print('replace?', animation_texture.path)
                    if fps != target_fps:
                        print(f'{fps} -> {target_fps}', '----------', animation_texture.path.name)
                        base_name = animation_texture.path.stem.split('_seed')[0]
                        seed = animation_texture.path.stem.split('_seed')[1].split('_',1)[0]
                        self.overwrite_dialog.enabled = True
                        self.overwrite_label.text = f'Overwrite\n{animation_texture.path.name}?'
                        self.overwrite_button.on_click = Sequence(
                            animation_texture.path.unlink,
                            Func(self.particle_system_container.bake, base_name, seed, target_fps),
                            self.overwrite_dialog.disable
                            )
                    return

            bake_button = Button(parent=self, scale=(.1,.05), text=f'Bake to VAT\n({target_fps}FPS)', text_size=.5, color=color.orange.tint(-.2))
            bake_button.on_click = _try_bake
            self.bake_buttons.append(bake_button)

        if self.play_baked_buttons:
            grid_layout(self.play_baked_buttons, max_x=2, offset=((window.aspect_ratio*-.5)+.25,.4), origin=(-.5,.5), spacing=(.01,.01))
        if self.bake_buttons:
            grid_layout(self.bake_buttons, max_x=3, offset=((window.aspect_ratio*-.5)+.25+.42,.4), origin=(-.5,.5), spacing=(.01,.01))




if __name__ == '__main__':
    app = Ursina()

    player = Entity(model='wireframe_cube', color=color.magenta, origin_y=-.5, alpha=1)
    run_particles = ParticleSystem(
        parent=player,
        scale=1,
        speed=1,
        spawn_interval=.05,
        num_particles=1,
        mesh='icosphere',
        world_space=True,
        end_color = color.red,
        end_size=0,
        direction_randomness=Vec3(360),
        loop_every=.1,
        auto_play=True
        # name='run particles',
    )
    EditorCamera()
    window.color = color.black

    # burst_particles = dict(position=player.position, scale=.5,
    #     speed=3,
    #     lifetime=.125,
    #     num_particles=6,
    #     direction_randomness=Vec3(360,360,360),
    #     move_directions='up',
    #     mesh=Cone(3, radius=.3),
    #     start_size=(1,1.5,1),
    #     end_size=0,
    #     size_curve = curve.linear,
    #     start_color = (color.gray, color.light_gray),
    #     end_color = (color.gray, color.light_gray),
    #     name='burst', seed=0,
    # )
    # power_up_particles = dict(position=player.position+Vec3(0,.5,0), scale=(.5,1.5,.5),
    #     speed=2,
    #     lifetime=.2,
    #     # direction_randomness=Vec3(360,360,360),
    #     move_directions='up',
    #     mesh='diamond',
    #     start_size=.5,
    #     end_size=.5,
    #     size_curve = curve.linear,
    #     start_color = (color.gray, color.light_gray),
    #     end_color = (color.gray, color.light_gray),

    #     num_particles=0,
    #     spawn_points = [v.xzy for v in Circle(radius=3).vertices][::-1],
    #     spawn_type = 'random',
    #     spawn_interval=.0125,
    #     # loop=True
    #     name='power_up_particles', seed=0,
    # )





    # landing_dust_particles = dict(position=player.position+Vec3(0,.5,0), scale=1,
    #     end_size = Vec3(0),
    #     size_curve=curve.linear,
    #     speed=4,
    #     speed_curve=curve.out_circ,
    #     lifetime=.75,
    #     direction_randomness=Vec3(0,360,0),
    #     spin=Vec3(0,15,0) * 10,
    #     # spin_curve=curve.linear,
    #     mesh='icosphere',
    #     start_color = (color.white, color.white),
    #     end_color = (color.light_gray, color.light_gray),
    #     color_curve=curve.out_expo,
    #     num_particles=10,
    #     spawn_type = 'burst',
    #     name='landing_dust', seed=0,

    #     shader=matcap_shader,
    #     texture='matcap_1',
    # )


    S = 5
    spawn_points = [Vec3(*[random.uniform(-S,S) for _ in range(3)]) for i in range(1)]
    print(spawn_points)
    particle_system_container = ParticleSystemContainer((
        ParticleSystem(
            # scale=.75 * .5,
            start_size=.5,
            end_size = Vec3(0),
            size_curve=curve.linear,
            speed=1,
            # speed_curve=curve.out_circ,
            lifetime=6,
            direction_randomness=Vec3(0,360,0),
            # spin=Vec3(0,15,0) * 10,
            # spin_curve=curve.linear,
            mesh='icosphere',
            start_color = (color.white, color.white),
            # end_color = (color.light_gray, color.light_gray),
            # color_curve=curve.out_expo,
            num_particles=10,
            spawn_type = 'random',
            name='snow',
            spawn_interval=.2,
            # seed=0,
            loop_every=1,
            spawn_points=spawn_points,
        ),
    ))




    # CAN_RETURN_BAKED = False
    # for particle_system_settings in (hit_impact_particles, burst_particles, power_up_particles, gold_particles, landing_dust_particles, jump_effects):
    #     if not isinstance(particle_system_settings, (tuple, list)):
    #         particle_system_settings = (particle_system_settings, )
    #     if any([e.get('loop_every', False) for e in particle_system_settings]):
    #         # skip looping particle system
    #         continue
    #     for e in particle_system_settings:
    #         e |= {'auto_play':False, }
    #         ParticleSystem(**e)

    particle_system_ui = ParticleSystemUI(Path(__file__), ParticleSystemContainer((run_particles, ), parent=player))
    # particle_system_ui.render_particle_system_list()


    # buttons = (hit_impact_button, burst_button, power_up_button, heal_button, gold_button, landing_dust_button, jump_effect_button)
    # buttons = []
    # for particle_system_setting in (hit_impact_particles, burst_particles, power_up_particles, heal_particles, gold_particles, landing_dust_particles, jump_effects, gems):
    #     if not isinstance(particle_system_setting, (tuple, list)):
    #         particle_system_setting = (particle_system_setting, )
    #     name = '/'.join([sub_system.get('name', 'unnamed') for sub_system in particle_system_setting])

    #     play_uncached_button = Button(scale=.1, text=f'{name}\n(uncached)', text_size=.5, color=color.orange)
    #     def play_uncached(particle_system_setting=particle_system_setting):
    #         for e in particle_system_setting:
    #             ParticleSystem(**e)




    #     if not isinstance(particle_system_setting, (tuple, list)):
    #         particle_system_setting = (particle_system_setting, )
    #     name = '/'.join([sub_system.get('name', 'unnamed') for sub_system in particle_system_setting])

    #     play_uncached_button = Button(scale=.1, text=f'{name}\n(uncached)', text_size=.5, color=color.orange)
    #     def play_uncached(particle_system_setting=particle_system_setting):
    #         for e in particle_system_setting:
    #             e |= {'use_cache':False, 'auto_play':True}
    #             ParticleSystem(**e)
    #     play_uncached_button.on_click = play_uncached
    #     buttons.append(play_uncached_button)

    #     play_button = Button(scale=.1, text=f'{name}\n', text_size=.5, color=color.azure)
    #     def play(particle_system_setting=particle_system_setting):
    #         for e in particle_system_setting:
    #             e |= {'use_cache':True, 'auto_play':True}
    #             ParticleSystem(**e)
    #     play_button.on_click = play
    #     buttons.append(play_button)

    #     for fps in (12, 30, 60):
    #         bake_button = Button(scale=.1, text=f'bake\n{name}\n{fps} fps', text_size=.5)
    #         def _bake(particle_system_setting=particle_system_setting, fps=fps):
    #             for e in particle_system_setting:
    #                 e |= {'use_cache':False, 'auto_play':False}
    #                 instance = ParticleSystem(**e)
    #                 instance.bake(fps=fps)
    #         bake_button.on_click = _bake
    #         buttons.append(bake_button)


    # grid_layout(buttons, max_x=5, offset=((window.aspect_ratio*-.5)+.25,.4), origin=(-.5,.5), spacing=(.01,.01))


    def update():
        h = max((held_keys['gamepad left stick x'], held_keys['d']-held_keys['a']), key=lambda x: abs(x))
        v = max((held_keys['gamepad left stick y'], held_keys['w']-held_keys['s']), key=lambda x: abs(x))
        move_speed = 5
        input_direction = Vec3(h,0,v).normalized()
        input_strength = min(Vec3(h,0,v).length(), 1)
        player.look_at_xz(player.position + input_direction)
        player.position += player.forward * time.dt * move_speed * input_strength
        run_particles.ignore = input_strength < .01


    # ParticleSystemUI()
    ground = Entity(model='plane', scale=8, texture='grass', texture_scale=Vec2(1), color=color.dark_gray)
    # FrameAnimation3d('test_particles_', fps=30, loop=True, position=(4,1,0), color=color.azure)

    app.run()



from pathlib import Path

files = [
    Path('platformer_tutorial.py'),
    ]

for path in files:
    code_parts = []
    comment_parts = []

    with open(path, encoding='utf8') as f:
        comment_block = []
        code_block = []
        is_comment_block = False

        lines = f.readlines()
        y = 0 # track the nujmber for non-commented lines
        for i, line in enumerate(lines):
            if line.startswith('#'):
                if not is_comment_block:
                    is_comment_block = True
                elif code_block:
                    # print('--------------------end code block and add it to the list')
                    code_parts.append((y, ''.join(code_block)))
                    y += len(code_block)
                    code_block = []

            if is_comment_block and not line.startswith('#'):
                is_comment_block = False
                comment_parts.append(''.join([l.lstrip('#').lstrip(' ') for l in comment_block]))
                comment_block = []

            if is_comment_block:
                comment_block.append(line)
            else:
                code_block.append(line)


    # print('code:', code_parts)
    # print('comments:', comment_parts)
    # print('converted:', path)
    text = f'''
[← Back, documentation.html]

### {path.stem.title().replace('_',' ')}
'''
    for i, (comment, code) in enumerate(zip(comment_parts, code_parts)):
        # text += f'## {i}\n'
        text += comment
        text += '\n```\n'
        text += code[1].rstrip()
        text += '\n```\n\n'

    text += '## Result\n'
    text += '```\n'
    text += ''.join([e[1] for e in code_parts])
    text += '\n```'

    with open(f'{path.stem}.sswg', 'w', encoding='utf-8') as text_file:
        text_file.write(text)



from ursina import *

class EditorDraggable():

    def __init__(self):
        self.entity = None
        self.hover_color = color.color(0, 0, .9)
        self.click_color = color.color(190, .5, .9, .8)


    def input(self, key):
        if self.entity.hovered:
            if key == 'left mouse down':
                self.entity.color = self.click_color
            elif key == 'left mouse up':
                self.entity.color = self.hover_color


    def on_mouse_enter(self):
        self.original_color = self.entity.color
        self.entity.color = self.hover_color


    def on_mouse_exit(self):
        self.entity.color = self.original_color


if __name__ == '__main__':
    app = Ursina()
    e = Entity("test entity")
    e.model = 'quad'
    e.collision = True
    e.collider = 'box'
    e.add_script(EditorDraggable())
    app.run()



from ursina import *
from direct.actor.Actor import Actor

if __name__ == '__main__':
    app = Ursina()

    # application.asset_folder = Path('/home/poke/Downloads/')

    Actor('rn.glb')

    EditorCamera()
    app.run()


from ursina import *


class TransformField(Button):

    def __init__(self):
        super().__init__()
        self.name = 'transform_field'

        self.transform_type = 0
        self.axis = Vec3(1, 0, 0)
        self.target_entity = None
        self.move_entity = None


    @property
    def direction(self, value):
        return Vec3(value[0], value[1], value[2])


    def input(self, key):
        if self.hovered and key == 'left mouse down':
            self.move_entity = True

        if key == 'left mouse up':
            self.move_entity = False

            if self.hovered and mouse.delta_drag == (0,0):
                print('enter number')


    def update(self):
        if self.move_entity:
            if self.transform_type == 0:
                self.target_entity.position += self.axis * mouse.velocity[0]
            elif self.transform_type == 1:
                self.target_entity.rotation += self.axis * mouse.velocity[0]
            elif self.transform_type == 2:
                self.target_entity.scale += self.axis * mouse.velocity[0]


if __name__ == '__main__':
    app = Ursina()
    tf = TransformField()
    tf.target_entity = Entity('target')
    app.run()



from ursina import *

class SaveSceneButton():

    def __init__(self):
        self.ask_for_scene_name_menu = load_prefab('ask_for_scene_name_menu')
        self.ask_for_scene_name_menu.enabled = False


    def input(self, key):
        if self.entity.hovered and key == 'left mouse down' or key == 'control-s':
            if scene.entity.name == 'untitled_scene':
                self.ask_for_scene_name_menu.enabled = True
                self.ask_for_scene_name_menu.input_field.editing = True
            else:
                save_prefab(scene.entity, 'scenes')
                scene.has_changes = False
                print('saved scene:', 'scenes', scene_entity.name)



from ursina import *



app = Ursina()

window.color=color.black
e1 = Entity(model='plane', scale=10, collider='box', color=color.dark_gray)
e2 = Entity(model='cube', collider='box', color=color.dark_gray)

# sun = Entity(model='sphere', scale=.05, color=color.yellow, add_to_scene_entities=False)
camera.orthographic = True
ed = EditorCamera()


def input(key):
    if key == 'left mouse down':
        print('generate shadows')
        # for e in (e1, e2):
            # Entity(model='plane', world_scale=.05, parent=e, )
        # e = Entity(model='plane', scale=.05, position=mouse.world_point)
        m = Entity(model=Mesh(mode='point', thickness=.1))
        for y in range(64):
            for x in range(64):
                ray = raycast(camera.world_position + Vec2(x/4,y/4), camera.forward)
                print(ray.hit)
                if ray.hit:
                    m.model.vertices.append(ray.world_point)
                    # e = Entity(model='plane', scale=.2, position=ray.world_point)
                    # e.look_at(camera, 'up')

        m.model.set_render_mode_perspective(True)
        m.model.generate()






app.run()



from ursina import *

app = Ursina()
camera.orthographic = True


# building = Entity()

def make_building():
    # pass
    scene.clear()
#     # [destroy(e) ]
    # p = Entity(model=Cylinder(radius=radius, resolution=edges))
    # for i in range(0, 360, 360//edges):
    #     origin = Entity(parent=p, rotation_y = i)
    #     wall = Entity(parent=origin, model='cube', scale=(10,1,2), origin_z=-.5, origin_y=-.5, z=radius, color=color.orange.tint(random.uniform(-.2,.1)))
    #     column = Entity(parent=wall, model=Cylinder(8), x=.5, world_scale=(2,5,2), color=color.gold.tint(random.uniform(-.2,.1)))

    # line = Entity(parent=p, model='cube', origin_z=-.5, color=color.green)
    # line.look_at(column)
    # line.scale_z = distance(column.world_position, line.world_position)
    # line.rotation_y = 0

    floor = Entity(model=Cylinder(settings['edges'], radius=settings['radius']), color=color.gold)
    points = [v for v in floor.model.vertices if v[1] < .01 and v != (0,0,0)]
    for p in points:
        p = [round(e) for e in p]

    points = set(points)
    for p in points:
        Entity(
            model=Cylinder(8),
            scale=(settings['column_radius'],settings['column_height'],settings['column_radius']),
            color=color.azure,
            position=p,
            origin_z=.5
            )
    print(points)



settings = {
    'radius' : 10,
    'edges' : 5,
    'column_radius' : 3,
    'column_height' : 5,
    }

for i, (key, value) in enumerate(settings.items()):
    slider = Slider(.1, 20, default=value, step=.1, text=key, x=-.6, y=-.45+(i/20), name=key+'_slider', eternal=True)
    print(i)
    def modify():
        value = slider.value
        settings[key] = value
        make_building()

    slider.on_value_changed = modify

# b = Button('lol')
# print('.............', b.collider)
# def input(key):
#     if key == 'd':
#         print('lol')
#         # print(scene.entities)
#         # print()
#         scene.clear()
        # print('.............', b.collider)
        # print(scene.entities)
        # print()

EditorCamera()
make_building()
app.run()



from ursina import Entity, Vec3, camera, color, mouse


class Cursor(Entity):
    def __init__(self, **kwargs):
        super().__init__()
        self.parent = camera.ui
        self.texture = 'cursor'
        self.model = 'quad'
        self.color = color.light_gray
        # self.origin = (-.49, .49)
        self.scale *= .05
        self.render_queue = 1

        for key, value in kwargs.items():
            setattr(self, key, value)


    def update(self):
        self.position = Vec3(mouse.x, mouse.y, -100)




if __name__ == '__main__':
    from ursina import Button, Mesh, Ursina
    app = Ursina()
    Button('button').fit_to_text()
    camera.orthographic = True
    camera.fov = 100
    cursor =  Cursor(model=Mesh(vertices=[(-.5,0,0),(.5,0,0),(0,-.5,0),(0,.5,0)], triangles=[(0,1),(2,3)], mode='line', thickness=2), scale=.02)
    mouse.visible = False
    app.run()
